import pytest
from embiggen.utils import AbstractModel
from unittest import TestCase


class ClassForTestAbstractModelStocastic(AbstractModel):

    def __init__(self):
        super().__init__(42)

    @classmethod
    def is_stocastic(cls) -> bool:
        return True

    @classmethod 
    def can_use_edge_weights(cls) -> bool:
        return False

    @classmethod 
    def can_use_node_types(cls) -> bool:
        return False

    @classmethod 
    def can_use_edge_types(cls) -> bool:
        return False
    
    @classmethod
    def model_name(cls) -> str:
        return "test_method_name"
    
    @classmethod
    def library_name(cls) -> str:
        return "test_library_name"
    
    @classmethod
    def task_name(cls) -> str:
        return "test_task_name"
    
    @classmethod
    def can_use_edge_type_features(cls) -> bool:
        return False
    
    @classmethod
    def can_use_edge_features(cls) -> bool:
        return False

        
class ClassForTestAbstractModelNonStocastic(AbstractModel):

    def __init__(self):
        super().__init__()

    @classmethod
    def is_stocastic(cls) -> bool:
        return False

    @classmethod 
    def requires_edge_weights(cls) -> bool:
        return True

    @classmethod 
    def requires_node_types(cls) -> bool:
        return True

    @classmethod 
    def requires_edge_types(cls) -> bool:
        return True
    
    @classmethod
    def model_name(cls) -> str:
        return "test_method_name"
    
    @classmethod
    def library_name(cls) -> str:
        return "test_library_name"
    
    @classmethod
    def task_name(cls) -> str:
        return "test_task_name"
    
    @classmethod
    def can_use_edge_type_features(cls) -> bool:
        return False
    
    @classmethod
    def can_use_edge_features(cls) -> bool:
        return False

class TestAbstractModel(TestCase):

    def setUp(self):
        pass

    def test_not_implemented_methods(self):

        stocastic = ClassForTestAbstractModelStocastic()
        non_stocastic = ClassForTestAbstractModelNonStocastic()

        for method_name in (
            "smoke_test_parameters",
            "task_involves_edge_weights",
            "task_involves_topology",
            "is_topological",
            "task_involves_node_types",
            "task_involves_edge_types",
            "clone",
        ):
            with pytest.raises(NotImplementedError):
                stocastic.__getattribute__(method_name)()
            with pytest.raises(NotImplementedError):
                non_stocastic.__getattribute__(method_name)()

    def test_implemented_methods(self):

        stocastic = ClassForTestAbstractModelStocastic()
        non_stocastic = ClassForTestAbstractModelNonStocastic()

        for method_name in (
            "parameters",
        ):
            stocastic.__getattribute__(method_name)()
            non_stocastic.__getattribute__(method_name)()

    def test_find_available_models(self):
        self.assertEqual(len(AbstractModel.find_available_models("Walklets SkipGram", "Node Embedding")), 2)
        self.assertEqual(len(AbstractModel.find_available_models("Walklets CBOW", "Node Embedding")), 1)

        with pytest.raises(ValueError):
            AbstractModel.find_available_models(
                "Walklets Impossible",
                "Node Embedding"
            )

        with pytest.raises(ValueError):
            AbstractModel.find_available_models(
                "Walklets Impossible",
                "Edge Parapello"
            )


"""Module providing generic abstract model."""
from typing import Callable
from embiggen.utils.abstract_models.list_formatting import format_list
from typing import Dict, Any, Type, List, Optional
from dict_hash import Hashable, sha256
import pandas as pd
from userinput.utils import must_be_in_set
import inspect


def abstract_class(klass: Type["AbstractModel"]) -> Type["AbstractModel"]:
    """Simply adds a descriptor for meta-programming and nothing else."""
    return klass


def is_not_implemented(method: Callable) -> bool:
    """Returns whether this method contains a raise for not being implemented."""
    return "raise NotImplementedError" in inspect.getsource(method)


def is_implemented(method: Callable) -> bool:
    """Returns whether this method is implemented."""
    return not is_not_implemented(method)


@abstract_class
class AbstractModel(Hashable):
    """Class defining properties of a generic abstract model."""

    MODELS_LIBRARY: Dict[str, Dict[str, Dict[str, Type["AbstractModel"]]]] = {}

    def __init__(self, random_state: Optional[int] = None):
        """Create new abstract model.

        Parameters
        ---------------
        random_state: Optional[int] = None
            The random state to use if the model is stocastic.
        """
        super().__init__()
        if self.is_stocastic() and random_state is None:
            raise ValueError(
                f"The provided model is stocastic, yet no "
                f"random state was provided. Please do provide "
                f"a random state to the model {self.model_name()} "
                f"from library {self.library_name()} and task "
                f"{self.task_name()}."
            )
        if not self.is_stocastic() and random_state is not None:
            raise ValueError(
                "The provided model is not stocastic, yet a "
                f"random state of `{random_state}` was provided. "
                f"Please do not provide a random state to the model "
                f"{self.model_name()} from library {self.library_name()} "
                f"and task {self.task_name()}."
            )

        can_use_edge_weights = self.__getattribute__("can_use_edge_weights")
        requires_positive_edge_weights = self.__getattribute__(
            "requires_positive_edge_weights"
        )

        if (
            is_implemented(can_use_edge_weights)
            and not can_use_edge_weights()
            and is_implemented(requires_positive_edge_weights)
        ):
            raise ValueError(
                "We have found an useless method in the "
                f"class {self.__class__.__name__}, implementing method "
                f"{self.model_name()} from library {self.library_name()} "
                f"and task {self.task_name()}. "
                "It does not make sense to implement the "
                f"`requires_positive_edge_weights` method when the `can_use_edge_weights` "
                "always returns False, as it is already handled "
                "in the root abstract model class."
            )

        # Identify and resolve tautological implementations.
        for graph_property in (
            "edge_types",
            "node_types",
            "edge_weights",
            "edge_type_features",
            "edge_features",
        ):
            requires = f"requires_{graph_property}"
            requires_method = self.__getattribute__(requires)
            can_use = f"can_use_{graph_property}"
            can_use_method = self.__getattribute__(can_use)
            is_using = f"is_using_{graph_property}"

            if is_not_implemented(requires_method) and is_not_implemented(
                can_use_method
            ):
                raise ValueError(
                    "We have found a missing method implementation in the "
                    f"class {self.__class__.__name__}, implementing method "
                    f"{self.model_name()} from library {self.library_name()} "
                    f"and task {self.task_name()}. "
                    f"It is strictly necessary to implement either the `{requires}` "
                    f"method or the {can_use} method in order to adhere to the model "
                    "interface and facilitate the integration with the pipelines."
                )

            if is_implemented(requires_method) and requires_method():
                for method in (can_use, is_using):
                    if is_implemented(self.__getattribute__(method)):
                        raise ValueError(
                            "We have found an useless method in the "
                            f"class {self.__class__.__name__}, implementing method "
                            f"{self.model_name()} from library {self.library_name()} "
                            f"and task {self.task_name()}. "
                            "It does not make sense to implement the "
                            f"`{method}` method when the `{requires}` "
                            "always returns True, as it is already handled "
                            "in the root abstract model class."
                        )
            if is_implemented(can_use_method) and not can_use_method():
                for method in (requires, is_using):
                    if is_implemented(self.__getattribute__(method)):
                        raise ValueError(
                            "We have found an useless method in the "
                            f"class {self.__class__.__name__}, implementing method "
                            f"{self.model_name()} from library {self.library_name()} "
                            f"and task {self.task_name()}. "
                            "It does not make sense to implement the "
                            f"`{method}` method when the `{can_use}` "
                            "always returns False, as it is already handled "
                            "in the root abstract model class."
                        )

        self._random_state = random_state

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Return parameters to create a model with minimal configuration to test execution."""
        raise NotImplementedError(
            (
                "The `smoke_test_parameters` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters of the model."""
        if self._random_state is None:
            return dict()
        return dict(random_state=self._random_state)

    def into_smoke_test(self):
        """Creates new instance with smoke test parameters."""
        return self.__class__(**{**self.parameters(), **self.smoke_test_parameters()})

    @classmethod
    def requires_edge_weights(cls) -> bool:
        """Returns whether the model requires edge weights."""
        try:
            if not cls.can_use_edge_weights():
                return False
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `requires_edge_weights` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def task_involves_edge_weights(cls) -> bool:
        """Returns whether the model task involves edge weights."""
        raise NotImplementedError(
            (
                "The `task_involves_edge_weights` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def can_use_edge_weights(cls) -> bool:
        """Returns whether the model can optionally use edge weights."""
        try:
            if (
                is_implemented(cls.requires_edge_weights)
                and cls.requires_edge_weights()
            ):
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `can_use_edge_weights` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    def is_using_edge_weights(self) -> bool:
        """Returns whether the model is parametrized to use edge weights."""
        try:
            if self.requires_edge_weights():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `is_using_edge_weights` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {self.__name__}."
            )
        )

    @classmethod
    def requires_positive_edge_weights(cls) -> bool:
        """Returns whether the model requires positive edge weights."""
        try:
            if not cls.requires_edge_weights():
                return False
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `requires_positive_edge_weights` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def task_involves_topology(cls) -> bool:
        """Returns whether the model task involves topology."""
        raise NotImplementedError(
            (
                "The `task_involves_topology` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def is_topological(cls) -> bool:
        """Returns whether this embedding is based on graph topology."""
        raise NotImplementedError(
            (
                "The `is_topological` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def requires_node_types(cls) -> bool:
        """Returns whether the model requires node types."""
        try:
            if not cls.can_use_node_types():
                return False
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `requires_node_types` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def task_involves_node_types(cls) -> bool:
        """Returns whether the model task involves node types."""
        raise NotImplementedError(
            (
                "The `task_involves_node_types` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def can_use_node_types(cls) -> bool:
        """Returns whether the model can optionally use node types."""
        try:
            if cls.requires_node_types():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `can_use_node_types` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    def is_using_node_types(self) -> bool:
        """Returns whether the model is parametrized to use node types."""
        try:
            if self.requires_node_types():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `is_using_node_types` method must be implemented "
                "in the child classes of abstract model, but was not implemented "
                f"in the class {self.__class__.__name__} implementing the model {self.model_name()} "
                f"from the library {self.library_name()}."
            )
        )

    @classmethod
    def requires_edge_types(cls) -> bool:
        """Returns whether the model requires edge types."""
        try:
            if not cls.can_use_edge_types():
                return False
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `requires_edge_types` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def task_involves_edge_types(cls) -> bool:
        """Returns whether the model task involves edge types."""
        raise NotImplementedError(
            (
                "The `task_involves_edge_types` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        try:
            if cls.requires_edge_types():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `can_use_edge_types` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    def is_using_edge_types(self) -> bool:
        """Returns whether the model is parametrized to use edge types."""
        try:
            if self.requires_edge_types():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `is_using_edge_types` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {self.__name__}."
            )
        )

    @classmethod
    def requires_edge_type_features(cls) -> bool:
        """Returns whether the model requires edge types."""
        try:
            if not cls.can_use_edge_type_features():
                return False
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `requires_edge_type_features` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def can_use_edge_type_features(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        try:
            if cls.requires_edge_type_features():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `can_use_edge_type_features` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    def is_using_edge_type_features(self) -> bool:
        """Returns whether the model is parametrized to use edge types."""
        try:
            if self.requires_edge_type_features():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `is_using_edge_type_features` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {self.__name__}."
            )
        )
    
    @classmethod
    def requires_node_type_features(cls) -> bool:
        """Returns whether the model requires edge types."""
        try:
            if not cls.can_use_node_type_features():
                return False
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `requires_node_type_features` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def can_use_node_type_features(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        try:
            if cls.requires_node_type_features():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `can_use_node_type_features` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    def is_using_node_type_features(self) -> bool:
        """Returns whether the model is parametrized to use edge types."""
        try:
            if self.requires_node_type_features():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `is_using_node_type_features` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {self.__name__}."
            )
        )

    @classmethod
    def requires_edge_features(cls) -> bool:
        """Returns whether the model requires edge types."""
        try:
            if not cls.can_use_edge_features():
                return False
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `requires_edge_features` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def can_use_edge_features(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        try:
            if cls.requires_edge_features():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `can_use_edge_features` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    def is_using_edge_features(self) -> bool:
        """Returns whether the model is parametrized to use edge types."""
        try:
            if self.requires_edge_features():
                return True
        except (NotImplementedError, RecursionError):
            pass
        raise NotImplementedError(
            (
                "The `is_using_edge_features` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {self.__name__}."
            )
        )

    @classmethod
    def task_name(cls) -> str:
        """Returns the task for which this model is being used."""
        raise NotImplementedError(
            (
                "The `task_name` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def library_name(cls) -> str:
        """Returns library name of the model."""
        raise NotImplementedError(
            (
                "The `library_name` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    @classmethod
    def model_name(cls) -> str:
        """Returns model name of the model."""
        raise NotImplementedError(
            (
                "The `model_name` method must be implemented "
                "in the child classes of abstract model. "
                f"You should implement it in the class {cls.__name__}."
            )
        )

    def clone(self) -> Type["AbstractModel"]:
        """Returns copy of the current model."""
        raise NotImplementedError(
            (
                "The `clone` method must be implemented in "
                "the child classes of the abstract model."
            )
        )

    def consistent_hash(self) -> str:
        """Returns consistent hash describing the model."""
        return sha256(
            dict(
                **self.parameters(),
                model_name=self.model_name(),
                library_name=self.library_name(),
                task_name=self.task_name(),
            )
        )

    @staticmethod
    def is_available() -> bool:
        """Returns whether the model class is actually available in the user system."""
        return True

    @classmethod
    def is_stocastic(cls) -> bool:
        """Returns whether the model is stocastic and has therefore a random state."""
        raise NotImplementedError(
            (
                "The `is_stocastic` method must be implemented "
                "in the child classes of abstract model. "
                f"It was not implemented in the class {cls.__name__}."
            )
        )

    def set_random_state(self, random_state: int):
        """Returns whether the model is stocastic and has therefore a random state."""
        if not self.is_stocastic():
            raise ValueError(
                "It does not make sense to set the random state of a "
                "non-stocastic model."
            )
        self._random_state = random_state

    @staticmethod
    def get_task_data(model_name: str, task_name: str) -> Dict[str, Dict]:
        """Returns data relative to the registered model and task data."""
        if len(model_name) == 0:
            raise ValueError("The provided model name is empty.")

        # We check if the provided string is not an empty string.
        if len(task_name) == 0:
            raise ValueError("The provided task name is empty.")

        task_name = must_be_in_set(task_name, AbstractModel.MODELS_LIBRARY, "task name")

        model_name = must_be_in_set(
            model_name, AbstractModel.MODELS_LIBRARY[task_name], "model name"
        )

        # We retrieve the task data.
        return AbstractModel.MODELS_LIBRARY[task_name][model_name]

    @staticmethod
    def get_library_data(
        model_name: str, task_name: str, library_name: str
    ) -> Type["AbstractModel"]:
        """Returns model relative library, task and model name."""
        task_data = AbstractModel.get_task_data(model_name, task_name)

        # We check if the provided string is not an empty string.
        if len(library_name) == 0:
            raise ValueError("The provided library name is empty.")

        library_name = must_be_in_set(library_name, task_data.keys(), "library name")

        # We retrieve the library data.
        return task_data[library_name]

    @classmethod
    def get_model_from_library(
        cls,
        model_name: str,
        task_name: Optional[str] = None,
        library_name: Optional[str] = None,
    ) -> Type["AbstractModel"]:
        """Returns list of models implementations available for given task and model.

        Parameters
        -------------------
        model_name: str
            The name of the model to retrieve.
        task_name: Optional[str] = None
            The task that this implementation of the model should be able to do.
            If not provided, it will be returned the model if it has only a single
            possible task. If multiple tasks are available, an exception will
            be raised.
        library_name: Optional[str] = None
            The library from which to get the implementation of this model.
            If not provided, it will be returned the model if it has only a single
            possible library. If multiple librarys are available, an exception will
            be raised.
        """
        if task_name is None:
            try:
                task_name = cls.task_name()
            except NotImplementedError as exception:
                model_data_frame = get_models_dataframe()
                if model_name not in model_data_frame.model_name.values:
                    raise ValueError(
                        (
                            f"The requested model `{model_name}` is not available. "
                            "Please do provide a valid model name to resolve this ambiguity."
                        )
                    ) from exception
                task_name = (
                    model_data_frame[model_data_frame.model_name == model_name]
                    .iloc[0]
                    .task_name
                )

        task_data = AbstractModel.get_task_data(model_name, task_name)

        if library_name is None:
            library_names = list(task_data.keys())
            if len(library_names) == 1:
                library_name = library_names[0]
            elif "Ensmallen" in library_names:
                library_name = "Ensmallen"
            else:
                formatted_list = format_list(library_names)
                raise ValueError(
                    (
                        f"The requested model `{model_name}` is available for "
                        "multiple libraries and no specific library was requested, "
                        "so it is unclear which library you intend to execute. "
                        f"Specifically, the available libraries are {formatted_list}. "
                        "Please do provide a library name to resolve this ambiguity."
                    )
                )

        model_class = AbstractModel.get_library_data(
            model_name, task_name, library_name
        )

        # If the model is not available, we just
        # instantiate it to cause its helpful ModuleNotFound
        # exception raise, with the informations to help the user.
        if not model_class.is_available():
            model_class()

        # Otherwise if the model is available, we return
        # its class to let the user do whathever they want.
        return model_class

    @staticmethod
    def find_available_models(
        model_name: str, task_name: str
    ) -> List[Type["AbstractModel"]]:
        """Returns list of models implementations available for given task and model.

        Parameters
        -------------------
        model_name: str
            The name of the model to retrieve.
        task_name: str
            The task that this implementation of the model should be able to do.
        """
        return [
            model
            for model in AbstractModel.get_task_data(model_name, task_name).values()
            if model.is_available()
        ]

    @staticmethod
    def register(model_class: Type["AbstractModel"]):
        """Registers the provided model in the model library.

        Parameters
        ------------------
        model_class:  Type["AbstractModel"]
            The class to register.
        """
        task_name = model_class.task_name()
        model_name = model_class.model_name()
        # If this is the first model of its kind to be registered.
        if task_name not in AbstractModel.MODELS_LIBRARY:
            AbstractModel.MODELS_LIBRARY[task_name] = {}

        # We retrieve the data for the model to enrich it.
        # This is NOT a copy, but a reference to the same STATIC object.
        model_data = AbstractModel.MODELS_LIBRARY[task_name]

        if model_name not in model_data:
            model_data[model_name] = {}

        task_data = model_data[model_name]

        class_name = model_class.__name__

        library_name = model_class.library_name()
        if library_name not in task_data:
            task_data[library_name] = model_class
        else:
            raise ValueError(
                f"The provided model called `{model_name}` with class name "
                f"`{class_name}`, implemented using the {library_name} library "
                "was already previously registered as available for the "
                f"`{task_name}` task. This is an implementation issue, "
                "so if you are seeing this problem either you are trying "
                "to register a custom model or you have found an error in "
                "the Embiggen library. If you believe this to be the latter "
                "please do open an issue in the Embiggen repository."
            )


def get_model_metadata(model_class: Type[AbstractModel]):
    """Return meetadata for the given model."""
    try:
        return {
            "model_name": model_class.model_name(),
            "task_name": model_class.task_name(),
            "library_name": model_class.library_name(),
            "available": model_class.is_available(),
            "requires_node_types": model_class.requires_node_types(),
            "can_use_node_types": model_class.requires_node_types()
            or model_class.can_use_node_types(),
            "requires_edge_types": model_class.requires_edge_types(),
            "can_use_edge_types": model_class.requires_edge_types()
            or model_class.can_use_edge_types(),
            "requires_edge_type_features": model_class.requires_edge_type_features(),
            "can_use_edge_type_features": model_class.requires_edge_type_features()
            or model_class.can_use_edge_type_features(),
            "requires_edge_features": model_class.requires_edge_features(),
            "can_use_edge_features": model_class.requires_edge_features()
            or model_class.can_use_edge_features(),
            "requires_edge_weights": model_class.requires_edge_weights(),
            "can_use_edge_weights": model_class.requires_edge_weights()
            or model_class.can_use_edge_weights(),
            "requires_positive_edge_weights": model_class.requires_positive_edge_weights(),
        }
    except NotImplementedError as exception:
        raise NotImplementedError(
            "Some of the mandatory static methods were not "
            f"implemented in model class {model_class.__name__}. "
            f"The previous exception was: {str(exception)}"
        ) from exception


def get_models_dataframe() -> pd.DataFrame:
    """Returns dataframe with informations about available models."""
    return pd.DataFrame(
        [
            get_model_metadata(model_class)
            for tasks in AbstractModel.MODELS_LIBRARY.values()
            for libraries in tasks.values()
            for model_class in libraries.values()
        ]
    )


def get_available_models_for_node_embedding() -> pd.DataFrame:
    """Returns dataframe with informations about available models for node embedding."""
    df = get_models_dataframe()
    return df[(df.task_name == "Node Embedding") & df.available]


def get_available_models_for_edge_embedding() -> pd.DataFrame:
    """Returns dataframe with informations about available models for edge embedding."""
    df = get_models_dataframe()
    return df[(df.task_name == "Edge Embedding") & df.available]


def get_available_models_for_edge_prediction() -> pd.DataFrame:
    """Returns dataframe with informations about available models for edge prediction."""
    df = get_models_dataframe()
    return df[(df.task_name == "Edge Prediction") & df.available]


def get_available_models_for_edge_label_prediction() -> pd.DataFrame:
    """Returns dataframe with informations about available models for edge-label prediction."""
    df = get_models_dataframe()
    return df[(df.task_name == "Edge Label Prediction") & df.available]


def get_available_models_for_node_label_prediction() -> pd.DataFrame:
    """Returns dataframe with informations about available models for node-label prediction."""
    df = get_models_dataframe()
    return df[(df.task_name == "Node Label Prediction") & df.available]



"""Keras Sequence for Open-world assumption GCN."""
from typing import List, Optional, Tuple, Type, Union

import numpy as np
import tensorflow as tf
from ensmallen import Graph  # pylint: disable=no-name-in-module
from keras_mixed_sequence import Sequence

from embiggen.sequences.generic_sequences import EdgePredictionSequence
from embiggen.utils import AbstractEdgeFeature


class GCNEdgePredictionSequence(Sequence):
    """Keras Sequence for running Neural Network on graph edge prediction."""

    def __init__(
        self,
        graph: Graph,
        support: Graph,
        kernels: Optional[List[tf.SparseTensor]],
        batch_size: int,
        return_node_types: bool = False,
        return_edge_types: bool = False,
        return_node_ids: bool = False,
        return_edge_node_ids: bool = True,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None,
        use_edge_metrics: bool = False,
    ):
        """Create new Open-world assumption GCN training sequence for edge prediction.

        Parameters
        --------------------------------
        graph: Graph,
            The graph from which to sample the edges.
        support: Graph
            The graph to be used for the topological metrics.
        kernels: Optional[List[tf.SparseTensor]]
            The kernel to be used for the convolutions.
        batch_size: int
            The batch size to use.
        return_node_types: bool = False
            Whether to return the node types.
        return_edge_types: bool = False
            Whether to return the edge types.
        return_node_ids: bool = False
            Whether to return the node IDs.
            These are needed when an embedding layer is used.
        return_edge_node_ids: bool = True
            Whether to return the edge node IDs.
            These are needed when an edge feature extractor is used.
        node_features: List[np.ndarray]
            The node features to be used.
        node_type_features: Optional[List[np.ndarray]]
            The node type features to be used.
            For instance, these could be BERT embeddings of the
            description of the node types.
            When the graph has multilabel node types,
            we will average the features.
        edge_type_features: Optional[List[np.ndarray]]
            The edge type features to be used.
            For instance, these could be BERT embeddings of the
            description of the edge types.
            When the graph has multilabel edge types,
            we will average the features.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None,
            The edge features to be used.
        use_edge_metrics: bool = False
            Whether to return the edge metrics.
        """
        if not graph.has_edges():
            raise ValueError(
                f"An empty instance of graph {graph.get_name()} was provided!"
            )
        
        if edge_features is None:
            edge_features = []

        if not isinstance(edge_features, list):
            edge_features = [edge_features]

        if edge_type_features is None:
            edge_type_features = []
        
        if not isinstance(edge_type_features, list):
            edge_type_features = [edge_type_features]

        if support is None:
            support = graph

        if (
            return_node_types or
            len(node_type_features) > 0
        ) and graph.has_unknown_node_types():
            raise ValueError(
                f"The provided graph {graph.get_name()} "
                "contains unknown node types but node types "
                "have been requested for the sequence."
            )
        
        if (
            return_edge_types or
            len(edge_type_features) > 0
        ) and graph.has_unknown_edge_types():
            raise ValueError(
                f"The provided graph {graph.get_name()} "
                "contains unknown edge types but edge types "
                "have been requested for the sequence."
            )

        # We verify that the provided edge features are valid
        for edge_feature in edge_features:
            if not issubclass(type(edge_feature), AbstractEdgeFeature):
                raise NotImplementedError(
                    f"The provided edge feature of type {type(edge_feature)} "
                    "is not supported. Please provide an object of type "
                    "subclass of AbstractEdgeFeature."
                )

            if not edge_feature.is_fit():
                edge_feature.fit(support)

        if kernels is None:
            kernels = []

        if not isinstance(kernels, list):
            kernels = [kernels]

        for kernel in kernels:
            assert kernel is not None, "The provided kernel is None."

        self._kernels = kernels
        if node_features is None:
            node_features = []
        self._node_features = [
            node_feature.astype(np.float32)
            for node_feature in node_features
        ]

        # We need to reshape the node IDs into a column vector
        # so that they match exactly the shape expected by the
        # embedding layer of the model.
        self._node_ids = graph.get_node_ids().reshape(-1, 1) if return_node_ids and self.has_kernels() else None
        self._return_edge_node_ids = return_edge_node_ids

        if return_node_types or len(node_type_features) > 0:
            if graph.has_multilabel_node_types():
                maximal_multilabel_number = graph.get_maximum_multilabel_count()
                node_types = np.zeros((graph.get_number_of_nodes(), maximal_multilabel_number), dtype=np.int32)
                for node_id in range(graph.get_number_of_nodes()):
                    node_type: Optional[np.ndarray] = graph.get_node_type_ids_from_node_id(node_id)
                    if node_type is not None:
                        node_types[node_id, :len(node_type)] = node_type
            else:
                node_types = graph.get_single_label_node_type_ids()

            assert node_types.shape[0] == graph.get_number_of_nodes(), (
                f"The provided node types have a different number of rows "
                f"than the number of nodes in the graph. Expected {graph.get_number_of_nodes()} "
                f"but found {node_types.shape[0]}. This is an internal error, please "
                "open an issue at the Embiggen GitHub repository."
            )

        if len(node_type_features) > 0:
            if graph.has_multilabel_node_types():
                self._node_type_features = []
                minus_node_types = node_types - 1
                node_types_mask = node_types == 0
                for node_type_feature in node_type_features:
                    self._node_type_features.append(np.ma.array(
                        node_type_feature[minus_node_types],
                        mask=np.repeat(
                            node_types_mask,
                            node_type_feature.shape[1]
                        ).reshape((
                            *node_types.shape,
                            node_type_feature.shape[1]
                        ))
                    ).mean(axis=-2).data)
            else:
                self._node_type_features = [
                    node_type_feature[node_types]
                    for node_type_feature in node_type_features
                ]
        else:
            self._node_type_features = []

        self._edge_type_features = edge_type_features
        self._return_edge_types = return_edge_types

        self._sequence = EdgePredictionSequence(
            graph=graph,
            support=support,
            return_node_types=False,
            return_edge_types=self.return_edge_types() or self.has_edge_type_features(),
            use_edge_metrics=use_edge_metrics,
            batch_size=batch_size
        )

        # We need to reshape the node types, when they are a flat array,
        # to be of shape (batch_size, 1) so that they can have a shape
        # identical to the shape expected from the model. Such reshaping
        # procedure comes to virtually no cost as it is just a view of
        # the original array.
        if return_node_types and not graph.has_multilabel_node_types():
            node_types = node_types.reshape(-1, 1)

        self._node_types = node_types if return_node_types else None

        self._edge_features = edge_features

        self._current_index = 0
        super().__init__(
            sample_number=graph.get_number_of_directed_edges(),
            batch_size=batch_size
        )

    def use_edge_metrics(self) -> bool:
        """Return whether to use edge metrics."""
        return self._sequence.use_edge_metrics()
    
    def return_node_types(self) -> bool:
        """Return whether to return node types."""
        return self._sequence.return_node_types()
    
    def return_edge_types(self) -> bool:
        """Return whether to return edge types."""
        return self._return_edge_types
    
    def get_graph(self) -> Graph:
        """Return graph."""
        return self._sequence.get_graph()
    
    def get_support(self) -> Graph:
        """Return support graph."""
        return self._sequence.get_support()

    def get_kernels(self) -> List[tf.SparseTensor]:
        """Return kernels."""
        return self._kernels
    
    def has_kernels(self) -> bool:
        """Return whether we have kernels."""
        return len(self._kernels) > 0
    
    def return_edge_node_ids(self) -> bool:
        """Return whether to return edge node IDs."""
        return self._return_edge_node_ids
    
    def get_edge_type_features(self, edge_type_ids: np.ndarray) -> Tuple[np.ndarray]:
        """Return edge type features."""
        edge_type_features = []

        for edge_type_feature in self._edge_type_features:
            edge_type_features.append(
                edge_type_feature[edge_type_ids]
            )

        return tuple(edge_type_features)
    
    def has_edge_type_features(self) -> bool:
        """Return whether we have edge type features."""
        return len(self._edge_type_features) > 0
    
    def get_node_features(
        self,
        sources: Optional[np.ndarray]=None,
        destinations: Optional[np.ndarray]=None
    ) -> Tuple[np.ndarray]:
        """Return the node features associated with the provided node IDs.
        
        Parameters
        ---------------------------
        sources: Optional[np.ndarray],
            The source node IDs.
        destinations: Optional[np.ndarray],
            The destination node IDs.

        Implementative details
        ---------------------------
        When the source and destination node IDs are not provided,
        we return the node features associated with all node IDs,
        as is expected when dealing with whole graph convolutions.

        When the source and destination node IDs are provided,
        we MUST not be in a graph convolution scenario, and we
        need to check that the provided kernels are an empty list
        otherwise we find ourselves in an illegal state.
        Furthermore, we must check that the length of the provided
        source and destination node IDs is the same, otherwise
        we find ourselves in an illegal state.

        Raises
        ---------------------------
        ValueError,
            When the provided source and destination node IDs
            are not both None or both not None.
        ValueError,
            When the provided source and destination node IDs
            have different lengths.
        ValueError,
            When the provided source and destination node IDs
            are not None but the provided kernels are not an
            empty list.
        """
        if not self.has_kernels() and (sources is None or destinations is None):
            raise ValueError(
                "The provided source and destination node IDs "
                "must be provided when kernels are not provided."
            )
        if sources is None and destinations is None:
            return tuple([
                node_feature
                for node_feature in (
                    *self._kernels,
                    *self._node_features,
                    *self._node_type_features,
                    self._node_ids,
                    self._node_types,
                )
                if node_feature is not None
            ])
        if sources is None or destinations is None:
            raise ValueError(
                "The provided source and destination node IDs "
                "must be either both None or both not None."
            )
        if len(sources) != len(destinations):
            raise ValueError(
                "The provided source and destination node IDs "
                "must have the same length."
            )
        
        if len(self._kernels) > 0:
            raise ValueError(
                "The provided source and destination node IDs "
                "must be None when kernels are provided."
            )
        
        return tuple([
            node_feature[node_ids.flatten()]
            for node_ids in (sources, destinations)
            for node_feature in (
                *self._node_features,
                *self._node_type_features,
                # Note that we do not return the node IDs as they are not needed when
                # we are not running a whole graph convolution.
                # self._node_ids,
                self._node_types,
            )
            if node_feature is not None
        ])
    
    def get_edge_features_from_edge_node_ids(self, sources: np.ndarray, destinations: np.ndarray) -> Tuple[np.ndarray]:
        """Returns the edge features associated with the provided edge node IDs.
        
        Parameters
        ---------------------------
        sources: np.ndarray,
            The source node IDs.
        destinations: np.ndarray,
            The destination node IDs.
            
        Returns
        ---------------------------
        The edge features associated with the provided edge node IDs.
        
        """
        rasterized_edge_features = []
        for edge_feature in self._edge_features:
            for edge_feature_name, rasterized_edge_feature in edge_feature.get_edge_feature_from_edge_node_ids(
                support=self.get_support(),
                sources=sources,
                destinations=destinations
            ).items():
                if not isinstance(rasterized_edge_feature, np.ndarray):
                    raise ValueError(
                        f"The provided edge feature {edge_feature_name} "
                        f"from the edge feature extractor {edge_feature.get_feature_name()}"
                        f"is not a numpy array but a {type(rasterized_edge_feature)}."
                    )
                if rasterized_edge_feature.shape[0] != sources.shape[0]:
                    raise ValueError(
                        f"The provided edge feature {edge_feature_name} "
                        f"from the edge feature extractor {edge_feature.get_feature_name()}"
                        "has a different number of rows than the "
                        "number of edges in the graph. "
                        f"Expected {sources.shape[0]} but found {rasterized_edge_feature.shape[0]}."
                    )
                rasterized_edge_features.append(
                    rasterized_edge_feature.reshape(sources.shape[0], -1)
                )
        return tuple(rasterized_edge_features)

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        values = list(self._sequence[idx][0])

        sources = values[0]
        destinations = values[1]

        current_batch_size = sources.shape[0]

        edge_features: Tuple[np.ndarray] = self.get_edge_features_from_edge_node_ids(
            sources,
            destinations
        )

        edge_type_features = self.get_edge_type_features(values[2]) if self.has_edge_type_features() else []

        # If the edge types are NOT requested as the model does not have an edge type
        # embedding layer, we remove the edge type IDs from the values, since we have
        # retrieved them solely to index correctly the edge type features.
        if self.has_edge_type_features() and not self.return_edge_types():
            values[2] = None

        # We reshape both sources and destinations to be of shape
        # (batch_size, 1) so that they can be used as inputs for
        # the embedding layer.
        values[0] = values[0].reshape(-1, 1)
        values[1] = values[1].reshape(-1, 1)

        # We remove the edge node IDs if they are not requested.
        if not self._return_edge_node_ids:
            values[0] = None
            values[1] = None

        # If we have to return the edge types, we need to reshape
        # them to be of shape (batch_size, 1) so that they can be
        # used as inputs for the embedding layer.
        if self.return_edge_types():
            values[2] = values[2].reshape(-1, 1)

        # If necessary, we add the padding as the last batch may be
        # smaller than the required size (number of nodes).
        delta = self.batch_size - current_batch_size
        if delta > 0 and self.has_kernels():
            values = [
                np.pad(value, (0, delta) if len(value.shape)
                       == 1 else [(0, delta), (0, 0)])
                for value in values
                if value is not None
            ]
            edge_features = [
                np.pad(edge_feature, [(0, delta), (0, 0)])
                for edge_feature in edge_features
            ]
            edge_type_features = [
                np.pad(edge_type_feature, [(0, delta), (0, 0)])
                for edge_type_feature in edge_type_features
            ]

        return (tuple([
            value
            for value in (
                *values,
                *edge_features,
                *edge_type_features,
                *(self.get_node_features() if self.has_kernels() else self.get_node_features(
                    sources=sources,
                    destinations=destinations
                ))
            )
            if value is not None
        ]),)



"""Submodule with utilities for pytorch models."""
import torch

def validate_torch_device(device: str) -> str:
    """Validate and sanitize torch device name.
    
    Parameters
    ------------------------
    device: str
        The device to use with Torch.
        This can either be `cuda`, `gpu`, or `auto`.

    Raises
    ------------------------
    ValueError
        If the device is not either cpu or cuda.
    ValueError
        If cuda was requested but CUDA is not available.
    """

    if torch.cuda.is_available():
        cuda_comment = (
            "Your Torch installation does detect CUDA "
            "installed in your system. Do consider using `cuda` "
            "as device option as it may be the faster option."
        )
        if device == "auto":
            device = "cuda"
    else:
        cuda_comment = (
            "Your Torch installation is not "
            "currently able to detect any CUDA device."
        )
        if device == "auto":
            device = "cpu"
    
    if device not in ("cpu", "cuda"):
        raise ValueError(
            f"The provided torch device `{device}` is not a supported "
            "torch device. Currently, the supported torch devices are "
            f"cpu and cuda, or `auto` for auto-dispatching. {cuda_comment}"
        )
        
    if device == "cuda" and not torch.cuda.is_available():
        raise ValueError(
            f"{cuda_comment} You have provided as device `cuda`. "
            "Either use `cpu` or ensure you have a working GPU "
            "and CUDA is installed and has a version compatible "
            "with the version of Torch you have installed."
        )
    
    return device



"""Module providing adapter class making edge prediction possible in sklearn models."""
from sklearn.base import ClassifierMixin
from typing import Type, Union, List
from embiggen.utils.sklearn_utils import must_be_an_sklearn_classifier_model
from embiggen.edge_prediction.sklearn_like_edge_prediction_adapter import (
    SklearnLikeEdgePredictionAdapter,
)
from embiggen.utils.abstract_models import abstract_class


@abstract_class
class SklearnEdgePredictionAdapter(SklearnLikeEdgePredictionAdapter):
    """Class wrapping Sklearn models for running ."""

    def __init__(
        self,
        model_instance: Type[ClassifierMixin],
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
        training_unbalance_rate: float = 1.0,
        training_sample_only_edges_with_heterogeneous_node_types: bool = False,
        use_scale_free_distribution: bool = True,
        use_edge_metrics: bool = False,
        prediction_batch_size: int = 2**15,
        random_state: int = 42,
    ):
        """Create the adapter for Sklearn object.

        Parameters
        ----------------
        model_instance: Type[ClassifierMixin]
            The class instance to be adapted into edge prediction.
        edge_embedding_method: Union[List[str], str] = "Concatenate",
            The method(s) to use to compute the edges.
            If multiple edge embedding are provided, they
            will be Concatenated and fed to the model.
            The supported edge embedding methods are:
             * Hadamard: element-wise product
             * Sum: element-wise sum
             * Average: element-wise mean
             * L1: element-wise subtraction
             * AbsoluteL1: element-wise subtraction in absolute value
             * SquaredL2: element-wise subtraction in squared value
             * L2: element-wise squared root of squared subtraction
             * Concatenate: Concatenate of source and destination node features
             * Min: element-wise minimum
             * Max: element-wise maximum
             * L2Distance: vector-wise L2 distance - this yields a scalar
             * CosineSimilarity: vector-wise cosine similarity - this yields a scalar
        training_unbalance_rate: float = 1.0
            Unbalance rate for the training non-existing edges.
        training_sample_only_edges_with_heterogeneous_node_types: bool = False
            Whether to sample negative edges exclusively between nodes with different node types
            to generate the negative edges used during the training of the model.
            This can be useful when executing a bipartite edge prediction task.
        use_scale_free_distribution: bool = True
            Whether to sample the negative edges for the TRAINING of the model
            using a zipfian-like distribution that follows the degree distribution
            of the graph. This is generally useful, as these negative edges are less
            trivial to predict then edges sampled uniformely.
            We stringly advise AGAINST using uniform sampling.
        use_edge_metrics: bool = False
            Whether to use the edge metrics from traditional edge prediction.
            These metrics currently include:
            - Adamic Adar
            - Jaccard Coefficient
            - Resource allocation index
            - Preferential attachment
        prediction_batch_size: int = 2**15
            Batch size to use for the predictions.
            Since usually rendering a whole dense graph edge embedding is not
            feaseable in main memory, we chunk it into more digestable smaller
            batches of edges.
        random_state: int
            The random state to use to reproduce the training.

        Raises
        ----------------
        ValueError
            If the provided model_instance is not a subclass of `ClassifierMixin`.
        """
        must_be_an_sklearn_classifier_model(model_instance)
        # We want to mask the decorator class name
        self.__class__.__name__ = model_instance.__class__.__name__
        self.__class__.__doc__ = model_instance.__class__.__doc__

        super().__init__(
            model_instance=model_instance,
            edge_embedding_methods=edge_embedding_methods,
            training_unbalance_rate=training_unbalance_rate,
            training_sample_only_edges_with_heterogeneous_node_types=training_sample_only_edges_with_heterogeneous_node_types,
            use_scale_free_distribution=use_scale_free_distribution,
            use_edge_metrics=use_edge_metrics,
            prediction_batch_size=prediction_batch_size,
            random_state=random_state,
        )

    @classmethod
    def library_name(cls) -> str:
        """Return name of the model."""
        return "scikit-learn"


"""Unit test class for Node-label prediction pipeline."""
from tqdm.auto import tqdm
from unittest import TestCase
import pytest
import numpy as np
import pandas as pd
from embiggen.node_label_prediction import node_label_prediction_evaluation
from embiggen import (
    get_available_models_for_node_label_prediction,
    get_available_models_for_node_embedding,
)
from embiggen.embedders.ensmallen_embedders.degree_spine import DegreeSPINE
from ensmallen.datasets.kgobo import MIAPA
from ensmallen import Graph
from ensmallen.datasets.linqs import Cora, get_words_data
import shutil
import os
from embiggen.utils import AbstractEmbeddingModel
from embiggen.node_label_prediction.node_label_prediction_model import (
    AbstractNodeLabelPredictionModel,
)
from embiggen.feature_preprocessors import GraphConvolution


class TestEvaluateNodeLabelPrediction(TestCase):
    """Unit test class for Node-label prediction pipeline."""

    def setUp(self):
        """Setup objects for running tests on Node-label prediction pipeline class."""
        self._graph, self._data = get_words_data(Cora())
        self._number_of_holdouts = 2

    def test_evaluate_embedding_for_node_label_prediction(self):
        """Test graph visualization."""
        if os.path.exists("experiments"):
            shutil.rmtree("experiments")

        df = get_available_models_for_node_label_prediction()
        feature = DegreeSPINE(embedding_size=5)
        for (
            evaluation_schema
        ) in AbstractNodeLabelPredictionModel.get_available_evaluation_schemas():
            holdouts = node_label_prediction_evaluation(
                holdouts_kwargs={"train_size": 0.8},
                node_features=[feature, self._data],
                models=df.model_name,
                library_names=df.library_name,
                graphs=self._graph,
                verbose=False,
                evaluation_schema=evaluation_schema,
                number_of_holdouts=self._number_of_holdouts,
                smoke_test=True,
            )

        self.assertEqual(holdouts.shape[0], self._number_of_holdouts * 2 * df.shape[0])

        if os.path.exists("experiments"):
            shutil.rmtree("experiments")

    def test_node_label_prediction_models_apis(self):
        df = get_available_models_for_node_label_prediction()
        graph = self._graph.remove_singleton_nodes()
        red = self._graph.set_all_node_types("red")
        multilabel_graph = (
            Cora().remove_edge_weights().remove_edge_types() | red
        ).add_selfloops()
        binary_graph = (
            red | MIAPA().remove_edge_types().set_all_node_types("blue")
        ).add_selfloops()
        for g in (graph, multilabel_graph, binary_graph):
            node_features = DegreeSPINE(embedding_size=10).fit_transform(g)
            for model_name in tqdm(df.model_name, desc="Testing model APIs"):
                model = AbstractNodeLabelPredictionModel.get_model_from_library(
                    model_name
                )().into_smoke_test()

                if (
                    g.has_multilabel_node_types()
                    and not model.supports_multilabel_prediction()
                ):
                    with pytest.raises(ValueError):
                        model.fit(g, node_features=node_features)

                    continue
                model.fit(g, node_features=node_features)
                if model.library_name() in ("TensorFlow", "scikit-learn", "LightGBM", "CatBoost", "XGBoost"):
                    path = f"model_{model.library_name()}_{model.model_name()}_{model.task_name()}.pkl.gz"
                elif model.library_name() == "Ensmallen":
                    path = f"model_{model.library_name()}_{model.model_name()}_{model.task_name()}.json"
                else:
                    raise NotImplementedError(
                        f"The model {model.model_name()} from library {model.library_name()} "
                        f"for the task {model.task_name()} "
                        "is not currently covered by the test suite! "
                        "Please add the model to the test suite."
                    )
                if os.path.exists(path):
                    os.remove(path)
                model.dump(path)
                restored_model = model.load(path)
                
                restored_model_parameters = {
                    key: value
                    for key, value in restored_model.parameters().items()
                    if isinstance(value, (list, tuple, np.ndarray)) or pd.notna(value)
                }

                model_parameters = {
                    key: value
                    for key, value in model.parameters().items()
                    if isinstance(value, (list, tuple, np.ndarray)) or pd.notna(value)
                }

                self.assertEqual(
                    restored_model_parameters,
                    model_parameters
                )

                model.predict(g, node_features=node_features)
                model.predict_proba(g, node_features=node_features)

    def test_model_recreation(self):
        """Test graph visualization."""
        df = get_available_models_for_node_label_prediction()
        for _, row in df.iterrows():
            model = AbstractNodeLabelPredictionModel.get_model_from_library(
                model_name=row.model_name,
                task_name=AbstractNodeLabelPredictionModel.task_name(),
                library_name=row.library_name,
            )()
            try:
                AbstractNodeLabelPredictionModel.get_model_from_library(
                    model_name=row.model_name,
                    task_name=AbstractNodeLabelPredictionModel.task_name(),
                    library_name=row.library_name,
                )(**model.parameters())
            except Exception as e:
                raise ValueError(
                    f"Found an error in model {row.model_name} "
                    f"implemented in library {row.library_name}."
                ) from e

    def test_all_node_embedding_models_as_feature(self):
        """Test graph visualization."""
        df = get_available_models_for_node_embedding()
        bar = tqdm(
            df.iterrows(),
            total=df.shape[0],
            leave=False,
            desc="Testing embedding methods",
        )
        graph = Graph.generate_random_connected_graph(
            random_state=100,
            number_of_nodes=100,
            node_type="red",
        ) | Graph.generate_random_connected_graph(
            minimum_node_id=100,
            random_state=1670,
            number_of_nodes=100,
            node_type="blue",
        )

        complex_model = [
            "RotatE",
            "ComplEx",
        ]

        graph = graph.sort_by_decreasing_outbound_node_degree()

        for _, row in bar:
            if row.requires_edge_weights:
                continue

            if row.model_name in complex_model:
                continue

            bar.set_description(
                f"Testing {row.model_name} from library {row.library_name}"
            )

            node_label_prediction_evaluation(
                holdouts_kwargs=dict(train_size=0.8),
                models="Decision Tree Classifier",
                node_features=AbstractEmbeddingModel.get_model_from_library(
                    model_name=row.model_name,
                    library_name=row.library_name,
                    task_name="Node Embedding",
                )(),
                node_features_preprocessing_steps=GraphConvolution(),
                graphs=graph,
                number_of_holdouts=self._number_of_holdouts,
                evaluation_schema="Monte Carlo",
                verbose=False,
                smoke_test=True,
            )



"""Test to validate that the model GloVe works properly with graph walks."""
from unittest import TestCase

import pytest
from ensmallen.datasets.kgobo import CIO
from tqdm.auto import tqdm

from embiggen import get_available_models_for_node_embedding
from embiggen.embedders import HOPEEnsmallen, embed_graph
from embiggen.utils.abstract_models.abstract_embedding_model import \
    AbstractEmbeddingModel


class TestNodeEmbeddingPipeline(TestCase):
    """Unit test for model GloVe on graph walks."""

    def test_embedding_pipeline(self):
        """Test that embed pipeline works."""
        df = get_available_models_for_node_embedding()
        bar = tqdm(
            df.iterrows(),
            total=df.shape[0],
            leave=False,
            desc="Testing embedding methods"
        )
        for _, row in bar:
            if row.requires_edge_weights:
                graph_name = "Usair97"
                repository = "networkrepository"
            else:
                graph_name = "Cora"
                repository="linqs"

            bar.set_description(f"Testing {row.model_name} from {row.library_name}")

            embed_graph(
                graph_name,
                repository=repository,
                embedding_model=row.model_name,
                library_name=row.library_name,
                smoke_test=True
            )

    def test_hope_ensmallen(self):
        """Test that embed pipeline works."""
        graph_name = "CIO"
        repository="kgobo"

        with pytest.raises(ValueError):
            embed_graph(
                graph_name,
                repository=repository,
                embedding_model="HOPE",
                metric="Jaccard",
                root_node_name="Hallo!",
                embedding_size=5
            )
        
        with pytest.raises(ValueError):
            embed_graph(
                graph_name,
                repository=repository,
                embedding_model="HOPE",
                metric="Ancestors Jaccard",
                root_node_name=None,
                embedding_size=5
            )

        for metric in HOPEEnsmallen.get_available_metrics():
            if "ancestor" in metric.lower():
                root_node_name = CIO().get_node_name_from_node_id(0)
            else:
                root_node_name = None
            embed_graph(
                graph_name,
                repository=repository,
                embedding_model="HOPE",
                metric=metric,
                root_node_name=root_node_name,
                embedding_size=5
            )
    
    def test_model_recreation(self):
        """Test that embed pipeline works."""
        df = get_available_models_for_node_embedding()

        for _, row in df.iterrows():
            model = AbstractEmbeddingModel.get_model_from_library(
                model_name=row.model_name,
                task_name="Node Embedding",
                library_name=row.library_name
            )()
            parameters = model.parameters()
            try:
                second_model = AbstractEmbeddingModel.get_model_from_library(
                    model_name=row.model_name,
                    task_name="Node Embedding",
                    library_name=row.library_name,
                )(**parameters)
                for key, value in second_model.parameters().items():
                    self.assertEqual(parameters[key], value)
            except Exception as exception:
                raise ValueError(
                    f"Found an error in model {row.model_name} "
                    f"implemented in library {row.library_name}."
                ) from exception


"""Unit test class for GraphTransformer objects."""
import os
from unittest import TestCase

import numpy as np
import pandas as pd
import pytest
from ensmallen import Graph
from ensmallen.datasets.kgobo import CIO
from ensmallen.datasets.linqs import Cora
from tqdm.auto import tqdm

from embiggen import (get_available_models_for_edge_embedding,
                      get_available_models_for_edge_prediction,
                      get_available_models_for_node_embedding)
from embiggen.edge_prediction import (DecisionTreeEdgePrediction,
                                      GraphSAGEEdgePrediction,
                                      edge_prediction_evaluation)
from embiggen.edge_prediction.edge_prediction_ensmallen.perceptron import \
    PerceptronEdgePrediction
from embiggen.edge_prediction.edge_prediction_model import \
    AbstractEdgePredictionModel
from embiggen.embedders.ensmallen_embedders.degree_spine import DegreeSPINE
from embiggen.embedders.ensmallen_embedders.hyper_sketching import \
    HyperSketching
from embiggen.embedding_transformers import EdgeTransformer
from embiggen.feature_preprocessors import GraphConvolution
from embiggen.utils import AbstractEmbeddingModel
from .cached_tests import cache_or_store
from glob import glob


class TestEvaluateEdgePrediction(TestCase):
    """Unit test class for GraphTransformer objects."""

    def setUp(self):
        """Setup objects for running tests on GraphTransformer objects class."""
        self._graph = Cora().filter_from_names(
            max_node_degree=50,
        ).remove_singleton_nodes()
        self._graph_without_node_types = self._graph.remove_node_types()
        self._subgraph_of_interest = self._graph.filter_from_names(
            edge_type_names_to_keep=["Paper2Paper"]
        )
        self._number_of_holdouts = 2

    def test_evaluate_edge_prediction_with_multimodal_features(self):
        """Test edge prediction with edge and edge type features."""
        if cache_or_store([
            "./tests/test_edge_prediction_pipeline.py",
            "./embiggen/embedders/ensmallen_embedders/degree_spine.py",
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            *glob("./embiggen/edge_prediction/**/*.py", recursive=True),
            *glob("./embiggen/edge_prediction/*.py",),
            "./embiggen/utils/abstract_models/abstract_classifier_model.py",
        ]):
            return
        df = get_available_models_for_edge_prediction()

        df = df[
            df.can_use_edge_features &
            df.can_use_edge_type_features
        ]

        holdouts = edge_prediction_evaluation(
            holdouts_kwargs=dict(train_size=0.9),
            models=df.model_name,
            library_names=df.library_name,
            node_features=DegreeSPINE(embedding_size=5),
            edge_features=HyperSketching(),
            node_type_features=np.random.uniform(
                size=(self._graph.get_number_of_node_types(), 7)
            ),
            edge_type_features=np.random.uniform(
                size=(self._graph.get_number_of_edge_types(), 3)
            ),
            evaluation_schema="Connected Monte Carlo",
            graphs=self._graph,
            number_of_holdouts=self._number_of_holdouts,
            verbose=True,
            smoke_test=True,
        )
        self.assertEqual(
            holdouts.shape[0],
            self._number_of_holdouts*2*df.shape[0]
        )

    def test_model_recreation(self):
        if cache_or_store([
            "./tests/test_edge_prediction_pipeline.py",
            "./embiggen/embedders/ensmallen_embedders/degree_spine.py",
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            *glob("./embiggen/edge_prediction/**/*.py", recursive=True),
            *glob("./embiggen/edge_prediction/*.py",),
            "./embiggen/utils/abstract_models/abstract_classifier_model.py",
        ]):
            return
        df = get_available_models_for_edge_prediction()

        for _, row in df.iterrows():
            model = AbstractEdgePredictionModel.get_model_from_library(
                model_name=row.model_name,
                task_name=AbstractEdgePredictionModel.task_name(),
                library_name=row.library_name
            )()
            try:
                AbstractEdgePredictionModel.get_model_from_library(
                    model_name=row.model_name,
                    task_name=AbstractEdgePredictionModel.task_name(),
                    library_name=row.library_name
                )(**model.parameters())
            except Exception as exception:
                raise ValueError(
                    f"Found an error in model {row.model_name} "
                    f"implemented in library {row.library_name}."
                ) from exception

    def test_evaluate_edge_prediction(self):
        if cache_or_store([
            "./tests/test_edge_prediction_pipeline.py",
            "./embiggen/embedders/ensmallen_embedders/degree_spine.py",
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            *glob("./embiggen/edge_prediction/**/*.py", recursive=True),
            *glob("./embiggen/edge_prediction/*.py",),
            "./embiggen/utils/abstract_models/abstract_classifier_model.py",
        ]):
            return
        df = get_available_models_for_edge_prediction()
        holdouts = edge_prediction_evaluation(
            holdouts_kwargs=dict(train_size=0.9),
            models=df.model_name,
            library_names=df.library_name,
            node_features=DegreeSPINE(embedding_size=5),
            evaluation_schema="Connected Monte Carlo",
            graphs=[self._graph, self._graph_without_node_types],
            number_of_holdouts=self._number_of_holdouts,
            verbose=True,
            smoke_test=True,
        )
        self.assertEqual(
            holdouts.shape[0],
            self._number_of_holdouts*2*2*df.shape[0]
        )

    def test_evaluate_edge_prediction_with_subgraphs(self):
        if cache_or_store([
            "./tests/test_edge_prediction_pipeline.py",
            "./embiggen/embedders/ensmallen_embedders/degree_spine.py",
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            *glob("./embiggen/edge_prediction/**/*.py", recursive=True),
            *glob("./embiggen/edge_prediction/*.py",),
            "./embiggen/utils/abstract_models/abstract_classifier_model.py",
        ]):
            return
        df = get_available_models_for_edge_prediction()
        holdouts = edge_prediction_evaluation(
            holdouts_kwargs=dict(train_size=0.9),
            models=df.model_name,
            library_names=df.library_name,
            node_features=DegreeSPINE(embedding_size=5),
            evaluation_schema="Connected Monte Carlo",
            graphs=self._graph,
            number_of_holdouts=self._number_of_holdouts,
            verbose=True,
            smoke_test=True,
            subgraph_of_interest=self._subgraph_of_interest
        )
        self.assertEqual(
            holdouts.shape[0], self._number_of_holdouts*2*df.shape[0])

    def test_evaluate_edge_prediction_with_node_types_features(self):
        if cache_or_store([
            "./tests/test_edge_prediction_pipeline.py",
            "./embiggen/embedders/ensmallen_embedders/degree_spine.py",
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            *glob("./embiggen/edge_prediction/**/*.py", recursive=True),
            *glob("./embiggen/edge_prediction/*.py",),
            "./embiggen/utils/abstract_models/abstract_classifier_model.py",
        ]):
            return
        df = get_available_models_for_edge_prediction()
        holdouts = edge_prediction_evaluation(
            holdouts_kwargs=dict(train_size=0.9),
            models=df.model_name,
            library_names=df.library_name,
            node_features=DegreeSPINE(embedding_size=5),
            node_type_features=np.random.uniform(
                size=(self._graph.get_number_of_node_types(), 5)
            ),
            evaluation_schema="Connected Monte Carlo",
            graphs=self._graph,
            number_of_holdouts=self._number_of_holdouts,
            verbose=True,
            smoke_test=True
        )
        self.assertEqual(
            holdouts.shape[0],
            self._number_of_holdouts * 2 * df.shape[0]
        )

    def test_edge_prediction_models_apis(self):
        if cache_or_store([
            "./tests/test_edge_prediction_pipeline.py",
            "./embiggen/embedders/ensmallen_embedders/degree_spine.py",
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            *glob("./embiggen/edge_prediction/**/*.py", recursive=True),
            *glob("./embiggen/edge_prediction/*.py",),
            "./embiggen/utils/abstract_models/abstract_classifier_model.py",
        ]):
            return
        df = get_available_models_for_edge_prediction()
        graph = CIO().remove_singleton_nodes()
        multi_label_graph = graph.set_all_node_types("hu") | graph
        node_features = DegreeSPINE(embedding_size=10).fit_transform(graph)
        bar = tqdm(
            df.model_name,
            total=df.shape[0],
            leave=False,
        )
        for g in (multi_label_graph, graph):
            first_names = g.get_node_names()[:5]
            second_names = g.get_node_names()[5:10]
            for model_name in bar:
                bar.set_description(
                    f"Testing API of {model_name}"
                )
                model = AbstractEdgePredictionModel.get_model_from_library(
                    model_name
                )()
                parameters = {}
                if "use_edge_metrics" in model.parameters():
                    parameters["use_edge_metrics"] = True
                if "use_node_embedding" in model.parameters():
                    parameters["use_node_embedding"] = True
                if "use_node_type_embedding" in model.parameters():
                    parameters["use_node_type_embedding"] = True

                if "edge_embedding_methods" in model.parameters():
                    parameters["edge_embedding_methods"] = [
                        "Concatenate", "Hadamard", "Average"]
                model_class = AbstractEdgePredictionModel.get_model_from_library(
                    model_name
                )
                node_type_features = np.random.uniform(size=(
                    g.get_number_of_node_types(),
                    7
                ))
                model = model_class(
                    **{
                        **model_class.smoke_test_parameters(),
                        **parameters
                    },
                )
                model.fit(
                    g,
                    node_features=node_features,
                    node_type_features=node_type_features
                )

                if model.library_name() in ("TensorFlow", "scikit-learn", "LightGBM", "CatBoost", "XGBoost"):
                    path = "model.pkl.gz"
                elif model.library_name() == "Ensmallen":
                    path = "model.json"
                else:
                    raise NotImplementedError(
                        f"The model {model.model_name()} from library {model.library_name()} "
                        "is not currently covered by the test suite!"
                    )
                if os.path.exists(path):
                    os.remove(path)
                model.dump(path)
                restored_model = model.load(path)

                # Check that the restored model is the same as the original one
                # and that the parameters are the same.
                # We must remove the NaN values from the restored model
                # because they are by definition not equal to themselves.

                restored_model_parameters = {
                    key: value
                    for key, value in restored_model.parameters().items()
                    if isinstance(value, (list, tuple, np.ndarray)) or pd.notna(value)
                }

                model_parameters = {
                    key: value
                    for key, value in model.parameters().items()
                    if isinstance(value, (list, tuple, np.ndarray)) or pd.notna(value)
                }

                self.assertEqual(
                    restored_model_parameters,
                    model_parameters
                )

                with pytest.raises(NotImplementedError):
                    model.fit(
                        g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        edge_features=67
                    )
                with pytest.raises(NotImplementedError):
                    model.predict(
                        g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        edge_features=67
                    )
                with pytest.raises(NotImplementedError):
                    model.predict_proba(
                        g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        edge_features=67
                    )
                for return_predictions_dataframe in (True, False):
                    model.predict(
                        g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_proba(
                        g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_bipartite_graph_from_edge_node_ids(
                        graph=g,
                        source_node_ids=[0, 1, 2, 3],
                        destination_node_ids=[4, 5, 6, 7],
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_proba_bipartite_graph_from_edge_node_ids(
                        graph=g,
                        source_node_ids=[0, 1, 2, 3],
                        destination_node_ids=[4, 5, 6, 7],
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_bipartite_graph_from_edge_node_names(
                        graph=g,
                        source_node_names=first_names,
                        destination_node_names=second_names,
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_proba_bipartite_graph_from_edge_node_names(
                        graph=g,
                        source_node_names=first_names,
                        destination_node_names=second_names,
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_bipartite_graph_from_edge_node_prefixes(
                        graph=g,
                        source_node_prefixes=["OIO"],
                        destination_node_prefixes=["IAO", "CIO"],
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_proba_bipartite_graph_from_edge_node_prefixes(
                        graph=g,
                        source_node_prefixes=["OIO"],
                        destination_node_prefixes=["IAO", "CIO"],
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_clique_graph_from_node_ids(
                        graph=g,
                        node_ids=[0, 1, 2],
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_proba_clique_graph_from_node_ids(
                        graph=g,
                        node_ids=[0, 1, 2],
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_clique_graph_from_node_names(
                        graph=g,
                        node_names=first_names,
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_proba_clique_graph_from_node_names(
                        graph=g,
                        node_names=first_names,
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_clique_graph_from_node_prefixes(
                        graph=g,
                        node_prefixes=["OIO:"],
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_proba_clique_graph_from_node_prefixes(
                        graph=g,
                        node_prefixes=["OIO:"],
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_clique_graph_from_node_type_names(
                        graph=g,
                        node_type_names=["biolink:NamedThing"],
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )
                    model.predict_proba_clique_graph_from_node_type_names(
                        graph=g,
                        node_type_names=["biolink:NamedThing"],
                        support=g,
                        node_features=node_features,
                        node_type_features=node_type_features,
                        return_predictions_dataframe=return_predictions_dataframe
                    )

    def test_tree_with_cosine(self):
        if cache_or_store([
            "./tests/test_edge_prediction_pipeline.py",
            "./embiggen/embedders/ensmallen_embedders/degree_spine.py",
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            *glob("./embiggen/edge_prediction/**/*.py", recursive=True),
            *glob("./embiggen/edge_prediction/*.py",),
            "./embiggen/utils/abstract_models/abstract_classifier_model.py",
        ]):
            return
        graph = CIO().remove_singleton_nodes().sort_by_decreasing_outbound_node_degree()
        for edge_embedding in EdgeTransformer.methods:
            for evaluation_schema in AbstractEdgePredictionModel.get_available_evaluation_schemas():
                model = DecisionTreeEdgePrediction(
                    edge_embedding_methods=edge_embedding
                )
                holdouts = edge_prediction_evaluation(
                    holdouts_kwargs=dict(train_size=0.9),
                    models=DecisionTreeEdgePrediction(
                        **{
                            **model.parameters(),
                            **model.smoke_test_parameters(),
                        }
                    ),
                    node_features=DegreeSPINE(embedding_size=10),
                    node_features_preprocessing_steps=GraphConvolution(
                        concatenate_features=True
                    ),
                    evaluation_schema=evaluation_schema,
                    graphs=graph,
                    number_of_holdouts=self._number_of_holdouts,
                    verbose=True,
                    validation_unbalance_rates=(1.0, 2.0,),
                )
                self.assertEqual(
                    holdouts.shape[0], self._number_of_holdouts*2*2)
                self.assertTrue(set(holdouts.validation_unbalance_rate)
                                == set((1.0, 2.0)))

    def test_all_node_embedding_models_as_feature(self):
        if cache_or_store([
            "./tests/test_edge_prediction_pipeline.py",
            "./embiggen/embedders/ensmallen_embedders/degree_spine.py",
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            *glob("./embiggen/edge_prediction/**/*.py", recursive=True),
            *glob("./embiggen/embedders/**/*.py", recursive=True),
            *glob("./embiggen/edge_prediction/*.py",),
            "./embiggen/utils/abstract_models/abstract_classifier_model.py",
        ]):
            return
        df = get_available_models_for_node_embedding()
        bar = tqdm(
            df.iterrows(),
            total=df.shape[0],
            leave=False,
            desc="Testing embedding methods"
        )

        complex_model = [
            "RotatE",
            "ComplEx",
        ]

        for _, row in bar:
            if row.requires_node_types:
                continue
            if row.requires_edge_weights:
                graph = Graph.generate_random_connected_graph(
                    random_state=100,
                    number_of_nodes=100,
                    weight=56.0,
                    edge_type="red"
                ) | Graph.generate_random_connected_graph(
                    random_state=1670,
                    number_of_nodes=100,
                    weight=526.0,
                    edge_type="red"
                )
            else:
                graph = CIO()

            if row.model_name in complex_model:
                continue

            bar.set_description(
                f"Testing {row.model_name} from library {row.library_name}")

            embedding_model = AbstractEmbeddingModel.get_model_from_library(
                model_name=row.model_name,
                library_name=row.library_name,
                task_name="Node Embedding"
            )()

            if embedding_model.requires_nodes_sorted_by_decreasing_node_degree():
                continue

            edge_prediction_evaluation(
                holdouts_kwargs=dict(train_size=0.9),
                models=PerceptronEdgePrediction(
                    edge_embeddings="CosineSimilarity"),
                node_features=embedding_model,
                evaluation_schema="Connected Monte Carlo",
                node_features_preprocessing_steps=GraphConvolution(),
                graphs=graph.remove_singleton_nodes(),
                number_of_holdouts=self._number_of_holdouts,
                verbose=False,
                smoke_test=True,
            )

    def test_all_edge_embedding_methods(self):
        if cache_or_store([
            "./tests/test_edge_prediction_pipeline.py",
            "./embiggen/embedders/ensmallen_embedders/degree_spine.py",
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            *glob("./embiggen/edge_prediction/**/*.py", recursive=True),
            *glob("./embiggen/edge_prediction/*.py",),
            "./embiggen/utils/abstract_models/abstract_classifier_model.py",
        ]):
            return
        if not GraphSAGEEdgePrediction.is_available():
            return
        for edge_embedding_method in GraphSAGEEdgePrediction.get_available_edge_embedding_methods():
            edge_prediction_evaluation(
                holdouts_kwargs=dict(train_size=0.9),
                models=GraphSAGEEdgePrediction(
                    edge_embedding_methods=edge_embedding_method,
                    epochs=1
                ),
                evaluation_schema="Connected Monte Carlo",
                graphs=self._graph,
                node_features="Degree-based SPINE",
                node_features_preprocessing_steps=GraphConvolution(),
                number_of_holdouts=self._number_of_holdouts,
                verbose=False
            )

    def test_all_edge_embedding_models_as_feature(self):
        """Test graph visualization."""
        if cache_or_store([
            "./tests/test_edge_prediction_pipeline.py",
            "./embiggen/embedders/ensmallen_embedders/degree_spine.py",
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            *glob("./embiggen/edge_prediction/**/*.py", recursive=True),
            *glob("./embiggen/edge_prediction/*.py",),
            "./embiggen/utils/abstract_models/abstract_classifier_model.py",
        ]):
            return
        df = get_available_models_for_edge_embedding()
        bar = tqdm(
            df.iterrows(),
            total=df.shape[0],
            leave=False,
            desc="Testing edge embedding methods"
        )
        for _, row in bar:
            if row.requires_node_types:
                continue
            if row.requires_edge_weights:
                graph = Graph.generate_random_connected_graph(
                    random_state=100,
                    number_of_nodes=100,
                    weight=56.0,
                    edge_type="red"
                ) | Graph.generate_random_connected_graph(
                    random_state=1670,
                    number_of_nodes=100,
                    weight=526.0,
                    edge_type="red"
                )
            else:
                graph = CIO()

            bar.set_description(
                f"Testing {row.model_name} from library {row.library_name}")

            edge_prediction_evaluation(
                holdouts_kwargs=dict(train_size=0.8),
                models="Decision Tree Classifier",
                edge_features=AbstractEmbeddingModel.get_model_from_library(
                    model_name=row.model_name,
                    library_name=row.library_name,
                    task_name="Edge Embedding"
                )(),
                evaluation_schema="Monte Carlo",
                graphs=graph,
                number_of_holdouts=self._number_of_holdouts,
                verbose=False,
                smoke_test=True,
            )



"""Module providing adapter class making edge prediction possible in sklearn models."""
from typing import Type, List, Optional, Dict, Any, Union, Tuple
import numpy as np
import math
import compress_pickle
import copy
from ensmallen import Graph
from embiggen.sequences.generic_sequences import EdgePredictionSequence
from embiggen.embedding_transformers import EdgePredictionTransformer, GraphTransformer
from embiggen.edge_prediction.edge_prediction_model import AbstractEdgePredictionModel
from embiggen.utils.abstract_models import abstract_class
from embiggen.utils import AbstractEdgeFeature
from tqdm.auto import tqdm


@abstract_class
class SklearnLikeEdgePredictionAdapter(AbstractEdgePredictionModel):
    """Class wrapping Sklearn models for running ."""

    def __init__(
        self,
        model_instance,
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
        training_unbalance_rate: float = 1.0,
        training_sample_only_edges_with_heterogeneous_node_types: bool = False,
        use_scale_free_distribution: bool = True,
        use_edge_metrics: bool = False,
        prediction_batch_size: int = 2**15,
        random_state: int = 42
    ):
        """Create the adapter for Sklearn object.

        Parameters
        ----------------
        model_instance: "Model"
            The class instance to be adapted into edge prediction.
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
            The method(s) to use to compute the edges.
            If multiple edge embedding are provided, they
            will be Concatenated and fed to the model.
            The supported edge embedding methods are:
             * Hadamard: element-wise product
             * Sum: element-wise sum
             * Average: element-wise mean
             * L1: element-wise subtraction
             * AbsoluteL1: element-wise subtraction in absolute value
             * SquaredL2: element-wise subtraction in squared value
             * L2: element-wise squared root of squared subtraction
             * Concatenate: Concatenate of source and destination node features
             * Min: element-wise minimum
             * Max: element-wise maximum
             * L2Distance: vector-wise L2 distance - this yields a scalar
             * CosineSimilarity: vector-wise cosine similarity - this yields a scalar
        training_unbalance_rate: float = 1.0
            Unbalance rate for the training non-existing edges.
        training_sample_only_edges_with_heterogeneous_node_types: bool = False
            Whether to sample negative edges exclusively between nodes with different node types
            to generate the negative edges used during the training of the model.
            This can be useful when executing a bipartite edge prediction task.
        use_scale_free_distribution: bool = True
            Whether to sample the negative edges for the TRAINING of the model
            using a zipfian-like distribution that follows the degree distribution
            of the graph. This is generally useful, as these negative edges are less
            trivial to predict then edges sampled uniformely.
            We stringly advise AGAINST using uniform sampling.
        use_edge_metrics: bool = False
            Whether to use the edge metrics from traditional edge prediction.
            These metrics currently include:
            - Adamic Adar
            - Jaccard Coefficient
            - Resource allocation index
            - Preferential attachment
        prediction_batch_size: int = 2**15
            Batch size to use for the predictions.
            Since usually rendering a whole dense graph edge embedding is not
            feaseable in main memory, we chunk it into more digestable smaller
            batches of edges.
        random_state: int
            The random state to use to reproduce the training.

        Raises
        ----------------
        ValueError
            If the provided model_instance is not a subclass of `ClassifierMixin`.
        """
        super().__init__(random_state=random_state)
        self._model_instance = model_instance
        self._edge_embedding_methods = edge_embedding_methods
        self._training_unbalance_rate = training_unbalance_rate
        self._prediction_batch_size = prediction_batch_size
        self._use_edge_metrics = use_edge_metrics
        self._use_scale_free_distribution = use_scale_free_distribution
        self._training_sample_only_edges_with_heterogeneous_node_types = training_sample_only_edges_with_heterogeneous_node_types

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters used for this model."""
        return {
            "training_sample_only_edges_with_heterogeneous_node_types": self._training_sample_only_edges_with_heterogeneous_node_types,
            "edge_embedding_methods": self._edge_embedding_methods,
            "training_unbalance_rate": self._training_unbalance_rate,
            "prediction_batch_size": self._prediction_batch_size,
            "use_edge_metrics": self._use_edge_metrics,
            "use_scale_free_distribution": self._use_scale_free_distribution,
            **super().parameters()
        }

    def clone(self):
        """Return copy of self."""
        return copy.deepcopy(self)

    def _trasform_graph_into_edge_embedding(
        self,
        graph: Union[Graph, Tuple[np.ndarray]],
        node_features: List[np.ndarray],
        support: Graph,
        node_types: Optional[Union[Graph, List[Optional[List[str]]], List[Optional[List[int]]]]] = None,
        edge_types: Optional[np.ndarray] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None,
    ) -> np.ndarray:
        """Transforms the provided data into an Sklearn-compatible numpy array.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        node_features: List[np.ndarray]
            The node features to be used in the training of the model.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_types: Optional[Union[Graph, List[Optional[List[str]]], List[Optional[List[int]]]]] = None,
            List of node types whose embedding is to be returned.
            This can be either a list of strings, or a graph, or if the
            aligned_mapping is setted, then this methods also accepts
            a list of ints.
        edge_types: Optional[np.ndarray] = None
            The edge types to use.
        node_type_features: Optional[List[np.ndarray]] = None,
            The node type features to be used in the training of the model.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None,
            The edge features to be used in the training of the model.

        Warns
        ------------------
        If the node features are provided as a numpy array it will not be possible
        to check whether the nodes in the graphs are aligned with the features.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        assert self.is_using_edge_types() == (edge_types is not None)

        graph_transformer = GraphTransformer(
            methods=self._edge_embedding_methods,
            aligned_mapping=True,
            include_both_undirected_edges=False
        )

        if edge_features is None:
            edge_features = []

        if not isinstance(edge_features, list):
            edge_features = [edge_features]

        for edge_feature in edge_features:
            if not issubclass(type(edge_feature), AbstractEdgeFeature):
                raise NotImplementedError(
                    f"Edge features of type {type(edge_feature)} are not supported."
                    "We currently only support edge features of type AbstractEdgeFeature."
                )

        rasterized_edge_features: List[np.ndarray] = []

        for edge_feature in edge_features:
            if isinstance(graph, Graph):
                for rasterized_edge_feature in edge_feature.get_edge_feature_from_graph(
                    graph=graph,
                    support=support,
                ).values():
                    if not isinstance(rasterized_edge_feature, np.ndarray):
                        raise ValueError(
                            f"The provided edge feature {edge_feature} returned a "
                            f"feature of type {type(rasterized_edge_feature)} "
                            f"when running on a graph of type {type(graph)}."
                            "We currently only support numpy arrays."
                        )
                    rasterized_edge_features.append(rasterized_edge_feature)
            elif isinstance(graph, tuple):
                for rasterized_edge_feature in edge_feature.get_edge_feature_from_edge_node_ids(
                    support=support,
                    sources=graph[0],
                    destinations=graph[1],
                ).values():
                    if not isinstance(rasterized_edge_feature, np.ndarray):
                        raise ValueError(
                            f"The provided edge feature {edge_feature} returned a "
                            f"feature of type {type(rasterized_edge_feature)} "
                            f"when running on a graph of type {type(graph)}."
                            "We currently only support numpy arrays."
                        )
                    rasterized_edge_features.append(rasterized_edge_feature)
            else:
                raise NotImplementedError(
                    f"A graph of type {type(graph)} was provided."
                )

        if self._use_edge_metrics:
            if isinstance(graph, Graph):
                edge_metrics = support.get_all_edge_metrics(
                    normalize=True,
                    subgraph=graph,
                )
                if not isinstance(edge_metrics, np.ndarray):
                    raise ValueError(
                        f"The provided graph returned a "
                        f"feature of type {type(edge_metrics)} "
                        f"when running on a graph of type {type(graph)}. "
                        "We currently only support numpy arrays."
                    )
                rasterized_edge_features.append(edge_metrics)
            elif isinstance(graph, tuple):
                edge_metrics = support.get_all_edge_metrics_from_node_ids(
                    source_node_ids=graph[0],
                    destination_node_ids=graph[1],
                    normalize=True,
                )
                if not isinstance(edge_metrics, np.ndarray):
                    raise ValueError(
                        f"The provided graph returned a "
                        f"feature of type {type(edge_metrics)} "
                        f"when running on a graph of type {type(graph)}. "
                        "We currently only support numpy arrays."
                    )
                rasterized_edge_features.append(edge_metrics)
            else:
                raise NotImplementedError(
                    f"A graph of type {type(graph)} was provided."
                )

        graph_transformer.fit(
            node_features,
            node_type_feature=node_type_features,
            edge_type_features=edge_type_features
        )

        return graph_transformer.transform(
            graph=graph,
            node_types=node_types,
            edge_types=edge_types,
            edge_features=rasterized_edge_features
        )

    def _fit(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None,
    ):
        """Run fitting on the provided graph.

        Parameters
        --------------------
        graph: Graph
            The graph to run predictions on.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[List[np.ndarray]] = None
            The node features to use.
        node_type_features: Optional[List[np.ndarray]] = None
            The node type features to use.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        """
        lpt = EdgePredictionTransformer(
            methods=self._edge_embedding_methods,
            aligned_mapping=True,
            include_both_undirected_edges=False
        )

        lpt.fit(
            node_features,
            node_type_feature=node_type_features,
            edge_type_features=edge_type_features
        )

        if support is None:
            support = graph

        if edge_features is None:
            edge_features: List[Type[AbstractEdgeFeature]] = []

        for edge_feature in edge_features:
            if not issubclass(type(edge_feature), AbstractEdgeFeature):
                raise NotImplementedError(
                    f"Edge features of type {type(edge_feature)} are not supported."
                    "We currently only support edge features of type AbstractEdgeFeature."
                )

        negative_graph = graph.sample_negative_graph(
            number_of_negative_samples=int(
                math.ceil(graph.get_number_of_edges() *
                          self._training_unbalance_rate)
            ),
            only_from_same_component=True,
            random_state=self._random_state,
            sample_only_edges_with_heterogeneous_node_types=self._training_sample_only_edges_with_heterogeneous_node_types,
            use_scale_free_distribution=self._use_scale_free_distribution,
            sample_edge_types=len(edge_type_features) > 0
        )

        assert negative_graph.has_edges()

        rasterized_edge_features = []

        for edge_feature in edge_features:
            for positive_edge_features, negative_edge_features in zip(
                edge_feature.get_edge_feature_from_graph(
                    graph=graph,
                    support=support,
                ).values(),
                edge_feature.get_edge_feature_from_graph(
                    graph=negative_graph,
                    support=support,
                ).values()
            ):
                rasterized_edge_features.append(np.vstack((
                    positive_edge_features,
                    negative_edge_features
                )))

        if self._use_edge_metrics:
            rasterized_edge_features.append(np.vstack((
                support.get_all_edge_metrics(
                    normalize=True,
                    subgraph=graph,
                ),
                support.get_all_edge_metrics(
                    normalize=True,
                    subgraph=negative_graph,
                )
            )))

        self._model_instance.fit(*lpt.transform(
            positive_graph=graph,
            negative_graph=negative_graph,
            edge_features=rasterized_edge_features,
            shuffle=True,
            random_state=self._random_state
        ))

    def _predict(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None,
    ) -> np.ndarray:
        """Run prediction on the provided graph.

        Parameters
        --------------------
        graph: Graph
            The graph to run predictions on.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[List[np.ndarray]]
            The node features to use.
        node_type_features: Optional[List[np.ndarray]] = None
            The node type features to use.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        """
        return self.predict_proba(
            graph=graph,
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
        ) > 0.5

    def _predict_proba(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None,
    ) -> np.ndarray:
        """Run prediction on the provided graph.

        Parameters
        --------------------
        graph: Graph
            The graph to run predictions on.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[List[np.ndarray]] = None
            The node features to use.
        node_type_features: Optional[List[np.ndarray]] = None
            The node type features to use.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        """
        if not graph.has_edges():
            return np.array([])

        sequence = EdgePredictionSequence(
            graph=graph,
            support=support,
            # The node types are handled within the context of the graph transformer
            # and guarantee to be aligned with the node type features and the sampled
            # source and destination nodes.
            return_node_types=False,
            # The same is not possible for the edge types, and therefore we need to
            # handle them separately within the edge prediction sequence.
            return_edge_types=self.is_using_edge_types(),
            use_edge_metrics=False,
            batch_size=self._prediction_batch_size
        )

        def predict(features):
            if hasattr(self._model_instance, "predict_proba"):
                return self._model_instance.predict_proba(features)

            predictions = self._model_instance.predict(features).astype(np.int32)
            prediction_probabilities = np.zeros(
                (predictions.shape[0], len(self._model_instance.classes_)),
                dtype=np.float32,
            )
            prediction_probabilities[np.arange(predictions.size), predictions] = 1
            return prediction_probabilities

        prediction_probabilities = np.concatenate([
            predict(self._trasform_graph_into_edge_embedding(
                graph=(edges[0][0], edges[0][1]),
                support=support,
                node_features=node_features,
                node_types=graph,
                edge_types=edges[0][2] if self.is_using_edge_types() else None,
                node_type_features=node_type_features,
                edge_type_features=edge_type_features,
                edge_features=edge_features,
            ))
            for edges in tqdm(
                (sequence[i] for i in range(len(sequence))),
                total=len(sequence),
                dynamic_ncols=True,
                desc="Running edge predictions",
                leave=False
            )
        ])

        assert len(prediction_probabilities.shape) in (1, 2)

        # In the majority but not totality of sklearn models,
        # the predictions of binary models are returned as
        # a couple of vectors for the positive and negative class.
        if len(prediction_probabilities.shape) > 1 and prediction_probabilities.shape[1] > 1:
            prediction_probabilities = prediction_probabilities[:, 1]

        return prediction_probabilities

    @classmethod
    def can_use_edge_weights(cls) -> bool:
        """Returns whether the model can optionally use edge weights."""
        return False

    @classmethod
    def can_use_node_types(cls) -> bool:
        """Returns whether the model can optionally use node types."""
        return True
    
    @classmethod
    def requires_node_types(cls) -> bool:
        """Returns whether the model requires node types."""
        return False

    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        return True
    
    @classmethod
    def requires_edge_types(cls) -> bool:
        """Returns whether the model requires edge types."""
        return False

    @classmethod
    def load(cls, path: str):
        """Load a saved version of the model from the provided path.

        Parameters
        -------------------
        path: str
            Path from where to load the model.
        """
        return compress_pickle.load(path)

    def dump(self, path: str):
        """Dump the current model at the provided path.

        Parameters
        -------------------
        path: str
            Path from where to dump the model.
        """
        compress_pickle.dump(self, path)

    @classmethod
    def requires_edge_type_features(cls) -> bool:
        return False
    
    @classmethod
    def requires_edge_features(cls) -> bool:
        return False
    
    @classmethod
    def can_use_edge_type_features(cls) -> bool:
        return True
    
    @classmethod
    def can_use_edge_features(cls) -> bool:
        return True


"""Unit tests to identify quickly potentially grammatical errors in the code base."""

import os
from glob import glob

import pytest


def test_typos():
    """Test for typos in the code base."""
    paths = glob("./embigfgn/**/*.py", recursive=True)

    # List of common typos in the code base relative
    # to the correct spelling of graph-related terms
    # and other such oddities.
    # The dictionary is structured with the correct
    # spelling as the key and a list of common typos
    # as the value.
    typos = {"edge": ["edg", "egde"], "prediction": ["prediction prediction"], "concatenation": ["conatenation"]}

    for path in paths:
        with open(path, "r", encoding="utf-8") as f:
            for line in f:
                line = line.lower()
                for correct, incorrect in typos.items():
                    for typo in incorrect:
                        if typo in line:
                            raise ValueError(
                                f"Found typo '{typo}' in line '{line}' in file '{path}'."
                                f" Please correct this typo to '{correct}'."
                            )



import pytest
from embiggen.utils.sklearn_utils import (
    is_sklearn_classifier_model,
    must_be_an_sklearn_classifier_model
)
from sklearn.tree import DecisionTreeClassifier
from unittest import TestCase


class TestSklearnUtils(TestCase):

    def setUp(self):
        pass

    def test_sklearn_utils(self):
        self.assertTrue(
            is_sklearn_classifier_model(DecisionTreeClassifier())
        )
        self.assertFalse(
            is_sklearn_classifier_model(DecisionTreeClassifier)
        )
        self.assertFalse(
            is_sklearn_classifier_model(int)
        )
        self.assertFalse(
            is_sklearn_classifier_model(6)
        )
        must_be_an_sklearn_classifier_model(
            DecisionTreeClassifier()
        )
        with pytest.raises(ValueError):
            must_be_an_sklearn_classifier_model(
                DecisionTreeClassifier
            )
        with pytest.raises(ValueError):
            must_be_an_sklearn_classifier_model(
                67
            )



"""Validates that all model parameters can be normalized"""

from embiggen import get_available_models_for_node_embedding, get_available_models_for_edge_prediction
from embiggen.utils.normalize_kwargs import normalize_kwargs

from embiggen.utils.abstract_models.abstract_embedding_model import AbstractEmbeddingModel
from embiggen.edge_prediction.edge_prediction_model import AbstractEdgePredictionModel


def test_normalize_kwargs_for_node_embedding_models():
    """Test that all node embedding models parameters can be normalized."""
    df = get_available_models_for_node_embedding()

    for _, row in df.iterrows():
        model = AbstractEmbeddingModel.get_model_from_library(
            model_name=row.model_name,
            task_name="Node Embedding",
            library_name=row.library_name,
        )()

        AbstractEmbeddingModel.get_model_from_library(
            model_name=row.model_name,
            task_name="Node Embedding",
            library_name=row.library_name
        )(**normalize_kwargs(model, model.parameters()))

        AbstractEmbeddingModel.get_model_from_library(
            model_name=row.model_name,
            task_name="Node Embedding",
            library_name=row.library_name
        )(**normalize_kwargs(model, model.smoke_test_parameters()))


def test_normalize_kwargs_for_link_prediction_models():
    """Test that all link prediction models parameters can be normalized."""
    df = get_available_models_for_edge_prediction()

    for _, row in df.iterrows():
        model = AbstractEdgePredictionModel.get_model_from_library(
            model_name=row.model_name,
            task_name=AbstractEdgePredictionModel.task_name(),
            library_name=row.library_name
        )()

        AbstractEdgePredictionModel.get_model_from_library(
            model_name=row.model_name,
            task_name=AbstractEdgePredictionModel.task_name(),
            library_name=row.library_name
        )(**normalize_kwargs(model, model.parameters()))

        AbstractEdgePredictionModel.get_model_from_library(
            model_name=row.model_name,
            task_name=AbstractEdgePredictionModel.task_name(),
            library_name=row.library_name
        )(**normalize_kwargs(model, model.smoke_test_parameters()))
        


"""Module with Keras Sequences."""
from embiggen.sequences.tensorflow_sequences.node2vec_sequence import Node2VecSequence
from embiggen.sequences.tensorflow_sequences.edge_prediction_training_sequence import EdgePredictionTrainingSequence
from embiggen.sequences.tensorflow_sequences.gcn_edge_prediction_training_sequence import GCNEdgePredictionTrainingSequence
from embiggen.sequences.tensorflow_sequences.gcn_edge_prediction_sequence import GCNEdgePredictionSequence
from embiggen.sequences.tensorflow_sequences.gcn_edge_label_prediction_training_sequence import GCNEdgeLabelPredictionTrainingSequence
from embiggen.sequences.tensorflow_sequences.gcn_edge_label_prediction_sequence import GCNEdgeLabelPredictionSequence
from embiggen.sequences.tensorflow_sequences.siamese_sequence import SiameseSequence

__all__ = [
    "Node2VecSequence",
    "EdgePredictionTrainingSequence",
    "GCNEdgePredictionTrainingSequence",
    "GCNEdgePredictionSequence",
    "GCNEdgeLabelPredictionTrainingSequence",
    "GCNEdgeLabelPredictionSequence",
    "SiameseSequence",
]



"""Wrapper for NNSED model provided from the Karate Club package."""
from typing import Dict, Any
from karateclub.community_detection import NNSED
from embiggen.embedders.karateclub_embedders.abstract_karateclub_embedder import AbstractKarateClubEmbedder


class NNSEDKarateClub(AbstractKarateClubEmbedder):

    def __init__(
        self,
        embedding_size: int = 32,
        iterations: int = 10,
        noise: float = 10**-6,
        random_state: int = 42,
        ring_bell: bool = False,
        enable_cache: bool = False
    ):
        """Return a new NNSED embedding model.

        Parameters
        ----------------------
        embedding_size: int = 32
            Number of dimensions. Default is 32.
        noise: float = 10**-6
            Random noise for normalization stability. Default is 10**-6.
        random_state: int = 42
            Random state to use for the stocastic
            portions of the embedding algorithm.
        ring_bell: bool = False,
            Whether to play a sound when embedding completes.
        enable_cache: bool = False
            Whether to enable the cache, that is to
            store the computed embedding.
        """
        self._iterations = iterations
        self._noise = noise

        super().__init__(
            embedding_size=embedding_size,
            enable_cache=enable_cache,
            ring_bell=ring_bell,
            random_state=random_state
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns the parameters used in the model."""
        return dict(
            **super().parameters(),
            iterations=self._iterations,
            noise=self._noise
        )

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(
            **AbstractKarateClubEmbedder.smoke_test_parameters(),
            iterations=1,
        )

    def _build_model(self) -> NNSED:
        """Return new instance of the NNSED model."""
        return NNSED(
            dimensions=self._embedding_size,
            iterations=self._iterations,
            noise=self._noise,
            seed=self._random_state
        )

    @classmethod
    def model_name(cls) -> str:
        """Returns name of the model"""
        return "NNSED"

    def get_memberships(self):
        """Getting the cluster membership of nodes."""
        return self._model.get_memberships()

    @classmethod
    def requires_nodes_sorted_by_decreasing_node_degree(cls) -> bool:
        return False

    @classmethod
    def is_topological(cls) -> bool:
        return True

    @classmethod
    def can_use_edge_weights(cls) -> bool:
        """Returns whether the model can optionally use edge weights."""
        return False

    @classmethod
    def can_use_node_types(cls) -> bool:
        """Returns whether the model can optionally use node types."""
        return False

    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        return False




"""Submodule wrapping Gradient Boosting for edge label prediction."""
from typing import Dict, Any, Union, List
from sklearn.ensemble import GradientBoostingClassifier
from embiggen.edge_label_prediction.edge_label_prediction_sklearn.sklearn_edge_label_prediction_adapter import SklearnEdgeLabelPredictionAdapter


class GradientBoostingEdgeLabelPrediction(SklearnEdgeLabelPredictionAdapter):
    """Create wrapper over Sklearn Gradient Boosting classifier for edge label prediction."""

    def __init__(
        self,
        loss='log_loss',
        learning_rate=0.1,
        n_estimators=100,
        subsample=1.0,
        criterion='friedman_mse',
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.,
        max_depth=3,
        min_impurity_decrease=0.,
        init=None,
        max_features="sqrt",
        verbose=0,
        max_leaf_nodes=None,
        warm_start=False,
        validation_fraction=0.1,
        n_iter_no_change=None,
        tol=1e-4,
        ccp_alpha=0.0,
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
        use_edge_metrics: bool = False,
        random_state: int = 42
    ):
        """Create the Gradient Boosting for Edge Label Prediction."""
        self._loss=loss
        self._learning_rate=learning_rate
        self._n_estimators=n_estimators
        self._criterion=criterion
        self._min_samples_split=min_samples_split
        self._min_samples_leaf=min_samples_leaf
        self._min_weight_fraction_leaf=min_weight_fraction_leaf
        self._max_depth=max_depth
        self._init=init
        self._subsample=subsample
        self._max_features=max_features
        self._max_leaf_nodes=max_leaf_nodes
        self._min_impurity_decrease=min_impurity_decrease
        self._warm_start=warm_start
        self._validation_fraction=validation_fraction
        self._n_iter_no_change=n_iter_no_change
        self._tol=tol
        self._ccp_alpha=ccp_alpha
        super().__init__(
            GradientBoostingClassifier(
                loss=loss, learning_rate=learning_rate, n_estimators=n_estimators,
                criterion=criterion, min_samples_split=min_samples_split,
                min_samples_leaf=min_samples_leaf,
                min_weight_fraction_leaf=min_weight_fraction_leaf,
                max_depth=max_depth, init=init, subsample=subsample,
                max_features=max_features,
                random_state=random_state, verbose=verbose,
                max_leaf_nodes=max_leaf_nodes,
                min_impurity_decrease=min_impurity_decrease,
                warm_start=warm_start, validation_fraction=validation_fraction,
                n_iter_no_change=n_iter_no_change, tol=tol, ccp_alpha=ccp_alpha
            ),
            edge_embedding_methods=edge_embedding_methods,
            use_edge_metrics=use_edge_metrics,
            random_state=random_state
        )
    
    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(
            max_depth=1,
            n_estimators=1
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters used for this model."""
        return {
            **super().parameters(),
            **dict(
                loss=self._loss,
                learning_rate=self._learning_rate,
                n_estimators=self._n_estimators,
                criterion=self._criterion,
                min_samples_split=self._min_samples_split,
                min_samples_leaf=self._min_samples_leaf,
                min_weight_fraction_leaf=self._min_weight_fraction_leaf,
                max_depth=self._max_depth,
                init=self._init,
                subsample=self._subsample,
                max_features=self._max_features,
                max_leaf_nodes=self._max_leaf_nodes,
                min_impurity_decrease=self._min_impurity_decrease,
                warm_start=self._warm_start,
                validation_fraction=self._validation_fraction,
                n_iter_no_change=self._n_iter_no_change,
                tol=self._tol,
                ccp_alpha=self._ccp_alpha
            )
        }

    @classmethod
    def model_name(cls) -> str:
        return "Gradient Boosting Classifier"


"""Module to test the XGBoost models for node-label, edge-label and edge prediction."""
from ensmallen import Graph
from ensmallen.datasets.kgobo import MIAPA
from ensmallen.datasets.linqs import Cora, get_words_data

from embiggen.edge_label_prediction import XGBEdgeLabelPrediction
from embiggen.edge_prediction import XGBEdgePrediction
from embiggen.embedders.ensmallen_embedders.degree_spine import DegreeSPINE
from embiggen.embedders.ensmallen_embedders.hyper_sketching import \
    HyperSketching
from embiggen.node_label_prediction import XGBNodeLabelPrediction

from .cached_tests import cache_or_store


def test_xgb_node_label_prediction_smoke_test():
    """Smoke test for the XGBoost node-label prediction model."""
    if cache_or_store([
        "tests/test_xgboost.py",
        "embiggen/embedders/ensmallen_embedders/degree_spine.py",
        "embiggen/node_label_prediction/node_label_prediction_xgboost/xgboost_node_label_prediction.py",
        "embiggen/node_label_prediction/sklearn_like_node_label_prediction_adapter.py",
        "embiggen/node_label_prediction/node_label_prediction_model.py",
        "embiggen/utils/abstract_models/abstract_classifier_model.py",
    ]):
        return
    graph, data = get_words_data(Cora())
    graph = graph.remove_singleton_nodes()
    red = graph.set_all_node_types("red")
    binary_graph = (
        red | MIAPA().remove_edge_types().set_all_node_types("blue")
    ).add_selfloops()
    another_binary_graph = Graph.generate_random_connected_graph(
        random_state=100,
        number_of_nodes=100,
        node_type="red",
    ) | Graph.generate_random_connected_graph(
        minimum_node_id=100,
        random_state=1670,
        number_of_nodes=100,
        node_type="blue",
    )

    for g, data in (
        (graph, data),
        (binary_graph, None),
        (another_binary_graph, None),
    ):
        if data is None:
            data = DegreeSPINE().into_smoke_test().fit_transform(g)
        model = XGBNodeLabelPrediction().into_smoke_test()
        model.fit(g, node_features=data)
        model.predict(g, node_features=data)
        model.predict_proba(g, node_features=data)


def test_xgb_edge_prediction_smoke_test():
    """Smoke test for the XGBoost edge-label prediction model."""
    if cache_or_store([
        "tests/test_xgboost.py",
        "embiggen/embedders/ensmallen_embedders/degree_spine.py",
        "embiggen/edge_prediction/edge_prediction_xgboost/xgboost_edge_prediction.py",
        "embiggen/edge_prediction/sklearn_like_edge_prediction_adapter.py",
        "embiggen/edge_prediction/edge_prediction_model.py",
        "embiggen/utils/abstract_models/abstract_classifier_model.py",
    ]):
        return
    graph, data = get_words_data(Cora())
    graph = graph.remove_singleton_nodes()
    red = graph.set_all_node_types("red")
    multilabel_graph = (
        Cora().remove_edge_weights().remove_edge_types() | red
    ).add_selfloops()
    binary_graph = (
        red | MIAPA().remove_edge_types().set_all_node_types("blue")
    ).add_selfloops()
    another_binary_graph = Graph.generate_random_connected_graph(
        random_state=100,
        number_of_nodes=100,
        node_type="red",
    ) | Graph.generate_random_connected_graph(
        minimum_node_id=100,
        random_state=1670,
        number_of_nodes=100,
        node_type="blue",
    )

    for g, data in (
        (graph, data),
        (multilabel_graph, None),
        (binary_graph, None),
        (another_binary_graph, None),
    ):
        if data is None:
            data = DegreeSPINE().into_smoke_test().fit_transform(g)
        model = XGBEdgePrediction().into_smoke_test()
        model.fit(g, node_features=data, edge_features=HyperSketching())
        model.predict(g, node_features=data, edge_features=HyperSketching())
        model.predict_proba(g, node_features=data,
                            edge_features=HyperSketching())


def test_xgb_edge_label_prediction_smoke_test():
    """Smoke test for the XGBoost edge-label prediction model."""
    if cache_or_store([
        "tests/test_xgboost.py",
        "embiggen/embedders/ensmallen_embedders/degree_spine.py",
        "embiggen/edge_label_prediction/edge_label_prediction_xgboost/xgboost_edge_label_prediction.py",
        "embiggen/edge_label_prediction/sklearn_like_edge_label_prediction_adapter.py",
        "embiggen/edge_label_prediction/edge_label_prediction_model.py",
        "embiggen/utils/abstract_models/abstract_classifier_model.py",
    ]):
        return
    graph, data = get_words_data(Cora())
    red = graph.set_all_edge_types("red")
    blue = MIAPA().remove_singleton_nodes().set_all_edge_types("blue")
    binary = red | blue

    for g, data in ((binary, None),):
        if data is None:
            data = DegreeSPINE().into_smoke_test().fit_transform(g)
        model = XGBEdgeLabelPrediction().into_smoke_test()
        model.fit(g, node_features=data, edge_features=HyperSketching())
        model.predict(g, node_features=data, edge_features=HyperSketching())
        model.predict_proba(g, node_features=data,
                            edge_features=HyperSketching())



"""Submodule providing utilities to normalize parameters for modular models."""
from typing import List, Any, Union, Type, Optional


def normalize_model_list_parameter(
    candidate_list: Optional[Union[Any, List[Any]]],
    elements_number: int,
    object_type: Type,
    default_value: Any = None,
    can_be_empty: bool = False
) -> List[Any]:
    """Return given candidate list parameter normalized.

    Parameters
    --------------------------
    candidate_list: Union[Any, List[Any]],
        Candidate list to be validated and normalized.
    elements_number: int,
        Number of expand to expand the list to or expected list size.
    object_type: Type,
        Base object of the list.
    default_value: Any = None,
        The type to use when the provided candidate list object is None.
    can_be_empty: bool = False
        Whether the list can be empty.

    Raises
    --------------------------
    TypeError,
        If the given candidate list base type is not compatible with the provided object type.
    ValueError,
        If the given number of elements in the list does not match the provided value.
    ValueError,
        If the number of elements provided is not a strictly positive integer.

    Returns
    --------------------------
    Validated and normalized list of parameters.
    """
    # Validate the provided number of elements
    if not isinstance(elements_number, int) or not can_be_empty and elements_number < 1:
        raise ValueError(
            (
                "The provided amount of elements is not strictly positive "
                "integer value. Namely, the provided object was `{}`."
            ).format(elements_number)
        )

    # Use the default value if the provided one is None.
    if candidate_list is None:
        candidate_list = default_value

    # Normalize the provided candidate list to a list
    if not isinstance(candidate_list, list):
        # Verify that the value in the list is effectively of the expected type
        if not isinstance(candidate_list, object_type):
            raise ValueError(
                (
                    "Object in candidate list was expected to be "
                    "of type `{}`, but object of type `{}` was found."
                ).format(
                    object_type,
                    type(candidate_list)
                )
            )
        # Expand the scalar to a list
        candidate_list = [
            candidate_list
            for _ in range(elements_number)
        ]

    # Validate the shape of the list
    # This is a tautological step when the provided element
    # was a scalar and expanded at the previous step.
    if len(candidate_list) != elements_number:
        raise ValueError(
            (
                "The provided candidate list length `{}` does not "
                "match the expected candidate list length `{}`."
            ).format(
                len(candidate_list),
                elements_number
            )
        )

    # Check that all elements in the list match with the
    # expected type.
    for element in candidate_list:
        # Verify that the element in the list is effectively of the expected type
        if not isinstance(element, object_type):
            raise ValueError(
                (
                    "Object `{}` in candidate list was expected to be "
                    "of type `{}`, but object of type `{}` was found."
                ).format(
                    element,
                    object_type,
                    type(element)
                )
            )

    # Finally, returned the validated and sanitized candidate list
    return candidate_list


def normalize_model_ragged_list_parameter(
    candidate_ragged_list: Optional[Union[Any, List[Any], List[List[Any]]]],
    submodules_number: int,
    layers_per_submodule: List[int],
    object_type: Type,
    default_value: Any = None
) -> List[List[Any]]:
    """Return validated and normalized ragged list of parameters.

    Parameters
    --------------------------
    candidate_ragged_list: Union[Any, List[Any], List[List[Any]]],
        Candidate ragged list to be normalized and validated.
    submodules_number: int,
        Number of submodules.
    layers_per_submodule: List[int]
        Dimensions for the sub-lists in the ragged list.
    object_type: Type,
        Base object of the ragged list.
    default_value: Any = None,
        The type to use when the provided candidate list object is None.

    Raises
    --------------------------
    ValueError,
        If the number of elements provided is not a strictly positive integer.

    Returns
    --------------------------
    Validated and normalized ragged list.
    """
    # Validate the layers per submodule list
    # and submodule
    normalize_model_list_parameter(
        layers_per_submodule,
        submodules_number,
        int,
        default_value
    )
    # Expand the provided ragged list if it is a scalar
    if not isinstance(candidate_ragged_list, list):
        candidate_ragged_list = normalize_model_list_parameter(
            candidate_ragged_list,
            submodules_number,
            object_type,
            default_value
        )
    # Expand the provided list of elements if it is not a list of lists
    candidate_ragged_list = [
        normalize_model_list_parameter(
            element,
            layers_number,
            object_type,
            default_value
        )
        for element, layers_number in zip(
            candidate_ragged_list,
            layers_per_submodule
        )
    ]
    # Return normalized ragged list
    return candidate_ragged_list



"""Module providing abstract Node2Vec implementation."""
from typing import Dict, Any
from ensmallen import Graph
import pandas as pd
from userinput.utils import must_be_in_set
from ensmallen import models
from embiggen.utils.abstract_models import abstract_class
from embiggen.embedders.ensmallen_embedders.ensmallen_embedder import EnsmallenEmbedder
from embiggen.utils import EmbeddingResult, normalize_kwargs


@abstract_class
class Node2VecEnsmallen(EnsmallenEmbedder):
    """Abstract class for Node2Vec algorithms."""

    MODELS = {
        "DeepWalk CBOW": models.CBOW,
        "DeepWalk SkipGram": models.SkipGram,
        "DeepWalk GloVe": models.GloVe,
        "Node2Vec CBOW": models.CBOW,
        "Node2Vec SkipGram": models.SkipGram,
        "Node2Vec GloVe": models.GloVe,
        "Walklets CBOW": models.WalkletsCBOW,
        "Walklets SkipGram": models.WalkletsSkipGram,
        "Walklets GloVe": models.WalkletsGloVe,
    }

    def __init__(
        self,
        embedding_size: int = 100,
        random_state: int = 42,
        ring_bell: bool = False,
        enable_cache: bool = False,
        **model_kwargs: Dict
    ):
        """Create new abstract Node2Vec method.

        Parameters
        --------------------------
        embedding_size: int = 100
            Dimension of the embedding.
        random_state: int = 42
            The random state to reproduce the training sequence.
        ring_bell: bool = False,
            Whether to play a sound when embedding completes.
        enable_cache: bool = False
            Whether to enable the cache, that is to
            store the computed embedding.
        model_kwargs: Dict
            Further parameters to forward to the model.
        """
        model_name = must_be_in_set(self.model_name(), self.MODELS.keys(), "model name")
        self._model_kwargs = normalize_kwargs(self, model_kwargs)
        
        self._model = Node2VecEnsmallen.MODELS[model_name](
            embedding_size=embedding_size,
            random_state=random_state,
            **model_kwargs
        )

        super().__init__(
            embedding_size=embedding_size,
            enable_cache=enable_cache,
            ring_bell=ring_bell,
            random_state=random_state
        )

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(
            epochs=1,
            embedding_size=5,
            window_size=1,
            walk_length=4,
            max_neighbours=10,
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters of the model."""
        return dict(
            **super().parameters(),
            **self._model_kwargs
        )

    def _fit_transform(
        self,
        graph: Graph,
        return_dataframe: bool = True,
    ) -> EmbeddingResult:
        """Return node embedding."""
        node_embeddings = self._model.fit_transform(graph)

        if "CBOW" in self.model_name():
            node_embeddings = list(reversed(node_embeddings))

        if return_dataframe:
            node_names = graph.get_node_names()
            node_embeddings = [
                pd.DataFrame(
                    node_embedding,
                    index=node_names
                )
                for node_embedding in node_embeddings
            ]
        return EmbeddingResult(
            embedding_method_name=self.model_name(),
            node_embeddings=node_embeddings
        )

    @classmethod
    def requires_edge_weights(cls) -> bool:
        return False

    @classmethod
    def requires_positive_edge_weights(cls) -> bool:
        return True

    @classmethod
    def can_use_edge_weights(cls) -> bool:
        """Returns whether the model can optionally use edge weights."""
        return True

    def is_using_edge_weights(self) -> bool:
        """Returns whether the model is parametrized to use edge weights."""
        return True

    @classmethod
    def can_use_node_types(cls) -> bool:
        """Returns whether the model can optionally use node types."""
        return True

    def is_using_node_types(self) -> bool:
        """Returns whether the model is parametrized to use node types."""
        return "change_node_type_weight" in self._model_kwargs and self._model_kwargs["change_node_type_weight"] != 1.0

    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        return True

    def is_using_edge_types(self) -> bool:
        """Returns whether the model is parametrized to use edge types."""
        return "change_edge_type_weight" in self._model_kwargs and  self._model_kwargs["change_edge_type_weight"] != 1.0

    @classmethod
    def is_stocastic(cls) -> bool:
        """Returns whether the model is stocastic and has therefore a random state."""
        return True

    @classmethod
    def requires_node_types(cls) -> bool:
        return False

    @classmethod
    def requires_edge_types(cls) -> bool:
        return False


"""EdgeTransformer class to convert edges to edge embeddings."""
from typing import List, Optional, Union, Optional

import numpy as np
import pandas as pd
from ensmallen import express_measures
from userinput.utils import must_be_in_set

from embiggen.embedding_transformers.node_transformer import NodeTransformer


def get_hadamard_edge_embedding(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return Hadamard edge embedding of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the Hadamard edge embedding.
    """
    return np.multiply(
        source_node_embedding,
        destination_node_embedding
    )


def get_sum_edge_embedding(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return sum edge embedding of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the sum edge embedding.
    """
    return np.add(
        source_node_embedding,
        destination_node_embedding
    )


def get_average_edge_embedding(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return average edge embedding of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the average edge embedding.
    """
    return np.divide(
        get_sum_edge_embedding(
            source_node_embedding,
            destination_node_embedding
        ),
        2.0
    )


def get_l1_edge_embedding(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return L1 edge embedding of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the L1 edge embedding.
    """
    return np.subtract(
        source_node_embedding,
        destination_node_embedding
    )


def get_l1_norm_edge_embedding(
    edge_embedding: np.ndarray
) -> np.ndarray:
    """Return L1 norm for each edge embedding row.

    Parameters
    --------------------------
    edge_embedding: np.ndarray
        Numpy array with the edge embeddings.

    Returns
    --------------------------
    Numpy array with the L1 norm scalar scores.
    """
    assert edge_embedding.ndim == 2
    return np.abs(edge_embedding).sum(axis=1, keepdims=True)


def get_absolute_l1_edge_embedding(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return Absolute L1 edge embedding of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the Absolute L1 edge embedding.
    """
    return np.abs(
        get_l1_edge_embedding(
            source_node_embedding,
            destination_node_embedding
        )
    )


def get_squared_l2_edge_embedding(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return Squared L2 edge embedding of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the Squared L2 edge embedding.
    """
    return np.power(
        get_l1_edge_embedding(
            source_node_embedding,
            destination_node_embedding
        ),
        2.0
    )


def get_l2_norm_edge_embedding(
    edge_embedding: np.ndarray
) -> np.ndarray:
    """Return L2 norm for each edge embedding row.

    Parameters
    --------------------------
    edge_embedding: np.ndarray
        Numpy array with the edge embeddings.

    Returns
    --------------------------
    Numpy array with the L2 norm scalar scores.
    """
    assert edge_embedding.ndim == 2
    return np.sqrt(np.power(edge_embedding, 2.0).sum(axis=1, keepdims=True))


def get_l2_edge_embedding(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return L2 edge embedding of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the L2 edge embedding.
    """
    return np.sqrt(
        get_squared_l2_edge_embedding(
            source_node_embedding,
            destination_node_embedding
        )
    )


def get_l2_distance(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return L2 distance of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the L2 distance.
    """
    return np.sqrt(np.sum(np.power(
        source_node_embedding - destination_node_embedding,
        2.0
    ), axis=1)).reshape((-1, 1))


def get_cosine_similarity(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return cosine similarity of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the cosine similarity.
    """
    assert source_node_embedding.dtype == destination_node_embedding.dtype
    assert source_node_embedding.shape == destination_node_embedding.shape
    hadamard_product = source_node_embedding * destination_node_embedding
    norm = (
        get_l2_norm_edge_embedding(source_node_embedding) *
        get_l2_norm_edge_embedding(destination_node_embedding)
    )
    norm[norm < 1e-6] = 1e-6
    return np.sum(hadamard_product, axis=1, keepdims=True) / norm


def get_concatenate_edge_embedding(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return Concatenate edge embedding of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the Concatenate edge embedding.
    """
    return np.hstack((
        source_node_embedding,
        destination_node_embedding
    ))


def get_min_edge_embedding(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return min edge embedding of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the min edge embedding.
    """
    return np.min(
        [
            source_node_embedding,
            destination_node_embedding
        ],
        axis=0
    )


def get_max_edge_embedding(
    source_node_embedding: np.ndarray,
    destination_node_embedding: np.ndarray
) -> np.ndarray:
    """Return max edge embedding of the two nodes.

    Parameters
    --------------------------
    source_node_embedding: np.ndarray
        Numpy array with the embedding of the source node.
    destination_node_embedding: np.ndarray
        Numpy array with the embedding of the destination node.

    Returns
    --------------------------
    Numpy array with the max edge embedding.
    """
    return np.max(
        [
            source_node_embedding,
            destination_node_embedding
        ],
        axis=0
    )


class EdgeTransformer:
    """EdgeTransformer class to convert edges to edge embeddings."""

    methods = {
        "Hadamard": get_hadamard_edge_embedding,
        "Sum": get_sum_edge_embedding,
        "Average": get_average_edge_embedding,
        "L1": get_l1_edge_embedding,
        "AbsoluteL1": get_absolute_l1_edge_embedding,
        "SquaredL2": get_squared_l2_edge_embedding,
        "L2": get_l2_edge_embedding,
        "Concatenate": get_concatenate_edge_embedding,
        "Min": get_min_edge_embedding,
        "Max": get_max_edge_embedding,
        "L2Distance": get_l2_distance,
        "CosineSimilarity": get_cosine_similarity,
    }

    def __init__(
        self,
        methods: Union[List[str], str] = "Hadamard",
        aligned_mapping: bool = False,
    ):
        """Create new EdgeTransformer object.

        Parameters
        ------------------------
        method: Union[List[str], str] = "Hadamard",
            Method to use for the embedding.
            If None is used, we return instead the numeric tuples.
            If multiple edge embedding are provided, they
            will be Concatenated and fed to the model.
            The supported edge embedding methods are:
             * Hadamard: element-wise product
             * Sum: element-wise sum
             * Average: element-wise mean
             * L1: element-wise subtraction
             * AbsoluteL1: element-wise subtraction in absolute value
             * SquaredL2: element-wise subtraction in squared value
             * L2: element-wise squared root of squared subtraction
             * Concatenate: Concatenate of source and destination node features
             * Min: element-wise minimum
             * Max: element-wise maximum
             * L2Distance: vector-wise L2 distance - this yields a scalar
             * CosineSimilarity: vector-wise cosine similarity - this yields a scalar
        aligned_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        """
        self._transformer = NodeTransformer(
            aligned_mapping=aligned_mapping,
        )

        if not isinstance(methods, list):
            methods = [methods]

        normalized_methods = []

        for method_name in methods:
            if not isinstance(method_name, str):
                raise ValueError(
                    "The provided method name should be a string, but we got "
                    f"{type(method_name)} instead."
                )
            normalized_methods.append(must_be_in_set(
                method_name,
                set(self.methods.keys()),
                "edge embedding method"
            ))

        self._methods = [
            self.methods[method_name]
            for method_name in normalized_methods
        ]
        self._method_names = normalized_methods
        self._edge_type_features = []

    def fit(
        self,
        node_feature: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
        node_type_feature: Optional[Union[pd.DataFrame, np.ndarray,
                                          List[Union[pd.DataFrame, np.ndarray]]]] = None,
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray,
                                           List[Union[pd.DataFrame, np.ndarray]]]] = None,
    ):
        """Fit the model.

        Parameters
        -------------------------
        node_feature: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
            Node feature to use to fit the transformer.
        node_type_feature: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            Node type feature to use to fit the transformer.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            Edge type feature to use to fit the transformer.
        """
        if edge_type_features is None:
            edge_type_features = []

        if not isinstance(edge_type_features, list):
            edge_type_features = [edge_type_features]

        # It also a good idea to check whether the edge type features provided
        # have all the same shape, and whether they contain any NaN value, so
        # to catch them early rather than later.

        # We expect that the edge type features can either be NumPy arrays of
        # Pandas Dataframes. If they are pandas dataframes, we will be able
        # to use their index as a way to map eventual edge types ids provided
        # with the string name.
        for edge_type_feature in edge_type_features:
            if not isinstance(edge_type_feature, (pd.DataFrame, np.ndarray)):
                raise ValueError(
                    "The provided edge type features should be either Pandas Dataframes or "
                    f"Numpy arrays, but we got {type(edge_type_feature)} instead."
                )

            if isinstance(edge_type_feature, pd.DataFrame):
                if edge_type_feature.index.hasnans:
                    raise ValueError(
                        "The provided edge type features should "
                        "not have NaN values in their index."
                    )

                if edge_type_feature.index.has_duplicates:
                    raise ValueError(
                        "The provided edge type features should "
                        "not have duplicated values in their index."
                    )

                if pd.isna(edge_type_feature).any():
                    raise ValueError(
                        "The provided edge type features should "
                        "not have NaN values in their values."
                    )

            if isinstance(edge_type_feature, np.ndarray):
                if np.isnan(edge_type_feature).any():
                    raise ValueError(
                        "The provided edge type features should "
                        "not have NaN values in their values."
                    )

        self._edge_type_features = edge_type_features
        self._transformer.fit(
            node_feature=node_feature,
            node_type_feature=node_type_feature,
        )

    def has_edge_type_features(self) -> bool:
        """Return whether the transformer has edge type features."""
        return len(self._edge_type_features) > 0
    
    def has_node_type_features(self) -> bool:
        """Return whether the transformer has node type features."""
        return self._transformer.has_node_type_features()
    
    def is_aligned_mapping(self) -> bool:
        """Return whether the transformer has aligned mapping."""
        return self._transformer.is_aligned_mapping()

    def has_numpy_edge_type_features(self) -> bool:
        """Returns whether any of the edge type features provided is a numpy array."""
        return any([
            isinstance(edge_type_feature, np.ndarray)
            for edge_type_feature in self._edge_type_features
        ])

    def transform(
        self,
        sources: Union[List[str], List[int]],
        destinations: Union[List[str], List[int]],
        source_node_types: Optional[Union[List[Optional[List[str]]],
                                          List[Optional[List[int]]]]] = None,
        destination_node_types: Optional[Union[List[Optional[List[str]]],
                                               List[Optional[List[int]]]]] = None,
        edge_types: Optional[Union[List[str], List[int], np.ndarray]] = None,
        edge_features: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
    ) -> np.ndarray:
        """Return embedding for given edges using provided method.

        Parameters
        --------------------------
        sources:Union[List[str], List[int]]
            List of source nodes whose embedding is to be returned.
        destinations:Union[List[str], List[int]]
            List of destination nodes whose embedding is to be returned.
        source_node_types: Optional[Union[List[Optional[List[str]]], List[Optional[List[int]]]]] = None,
            List of source node types whose embedding is to be returned.
            This can be either a list of strings, or a graph, or if the
            aligned_mapping is setted, then this methods also accepts
            a list of ints.
        destination_node_types: Optional[Union[List[Optional[List[str]]], List[Optional[List[int]]]]] = None,
            List of destination node types whose embedding is to be returned.
            This can be either a list of strings, or a graph, or if the
            aligned_mapping is setted, then this methods also accepts
            a list of ints.
        edge_types: Optional[Union[List[str], List[int], np.ndarray]] = None
            List of edge types whose embedding is to be returned.
        edge_features: Optional[Union[np.ndarray, List[np.ndarray]]] = None
            Optional edge features to be used as input Concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the provided graph.

        Raises
        --------------------------
        ValueError
            If embedding is not fitted.
        ValueError
            If the edge features are provided and do not have the correct shape.

        Returns
        --------------------------
        Numpy array of embeddings.
        """
        if self.has_edge_type_features() and edge_types is None:
            raise ValueError(
                "The edge type features are provided, but no edge types are provided."
            )
        
        if self.has_node_type_features() and source_node_types is None:
            raise ValueError(
                "The node type features are provided, but no source node types are provided."
            )
        
        if len(sources) != len(destinations):
            raise ValueError(
                "The provided sources and destinations should have the same length, "
                f"but we got {len(sources)} and {len(destinations)} instead."
            )
        
        # The source, destination and edge types, if provided, must have the same length.
        if edge_types is not None and len(destinations) != len(edge_types):
            raise ValueError(
                "The provided sources, destinations and edge types should have the same length, "
                f"but we got {len(sources)}, {len(destinations)} and {len(edge_types)} instead."
            )

        edge_type_features: List[np.ndarray] = []

        for edge_type_feature in self._edge_type_features:
            if isinstance(edge_type_feature, pd.DataFrame):
                if isinstance(edge_types[0], str):
                    edge_type_features.append(
                        edge_type_feature.loc[edge_types].values
                    )
                elif isinstance(edge_types[0], (int, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64)):
                    edge_type_features.append(
                        edge_type_feature.iloc[edge_types].values
                    )
                else:
                    raise ValueError(
                        "The provided edge types should be either strings or integers, but we got "
                        f"{type(edge_types[0])} instead."
                    )
            elif isinstance(edge_type_feature, np.ndarray):
                if isinstance(edge_types[0], str):
                    raise ValueError(
                        "Since the edge type features are provided as numpy arrays, "
                        "the edge types should be provided as integers. "
                        f"We got instead {type(edge_types[0])}."
                    )
                elif isinstance(edge_types[0], (int, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64)):
                    edge_type_features.append(
                        edge_type_feature[edge_types]
                    )
                else:
                    raise ValueError(
                        "The provided edge types should be either strings or integers, but we got "
                        f"{type(edge_types[0])} instead."
                    )
            else:
                raise ValueError(
                    "The provided edge type features should be either Pandas Dataframes or "
                    f"Numpy arrays, but we got {type(edge_type_feature)} instead."
                )
            
        assert len(self._edge_type_features) == len(edge_type_features)
        
        edge_embeddings: List[np.ndarray] = []
        if self._transformer.is_fit():
            for method in self._methods:
                edge_embedding = method(
                    self._transformer.transform(
                        sources,
                        node_types=source_node_types
                    ),
                    self._transformer.transform(
                        destinations,
                        node_types=destination_node_types
                    )
                )
                assert not np.isnan(edge_embedding).any(), (
                    "The provided edge embedding should not have NaN values, but we got "
                    f"a numpy array with shape {edge_embedding.shape} and NaN values. "
                    f"The object was obtained using the method {method}."
                )
                edge_embeddings.append(edge_embedding)

        if edge_features is None:
            edge_features = []

        if not isinstance(edge_features, list):
            edge_features = [edge_features]

        for edge_feature in edge_features:
            if not isinstance(edge_feature, np.ndarray):
                raise ValueError(
                    "The provided edge features should be numpy arrays, but we got "
                    f"{type(edge_feature)} instead."
                )

            if len(edge_embeddings) > 0 and edge_feature.shape[0] != edge_embeddings[0].shape[0]:
                raise ValueError(
                    "The provided edge features should have a sample for each of the edges "
                    f"in the graph, which are {sources.shape[0]}, with embedding "
                    f"shape {edge_embeddings[0].shape}, but has shape {edge_feature.shape}."
                )

        if all([
            len(features) == 0
            for features in (
                edge_features,
                edge_type_features,
                edge_embeddings
            )
        ]):
            raise ValueError(
                "At least one of the provided edge features, edge type features or edge embeddings "
                "should be provided."
            )

        expected_shape: Optional[int] = None
        for features, feature_kind in (
            (edge_features, "edge features"),
            (edge_type_features, "edge type features"),
            (edge_embeddings, "edge embeddings")
        ):
            if len(features) > 0 and expected_shape is None:
                # We imprint the expected shape on the
                # first feature available.
                expected_shape = features[0].shape[0]
            # We check that all the features have the same first dimension.
            for feature in features:
                assert not np.isnan(feature).any(), (
                    "The provided edge features should not have NaN values, but we got "
                    f"a numpy array with shape {feature.shape} and NaN values. "
                    f"It is a {feature_kind}."
                )
                if feature.shape[0] != expected_shape:
                    raise ValueError(
                        "The provided edge features should have a sample for each of the edges "
                        f"in the graph, which are {expected_shape}, but we got {feature.shape[0]}."
                    )


        result = np.hstack([
            *[
                edge_embedding.reshape((expected_shape, -1))
                for edge_embedding in edge_embeddings
            ],
            *[
                edge_feature.reshape((expected_shape, -1))
                for edge_feature in edge_features
            ],
            *[
                edge_type_feature.reshape((expected_shape, -1))
                for edge_type_feature in edge_type_features
            ]
        ])

        return result
